import numpy as np

from .backend import get_backend
from .validation import check_random_state


def compute_lipschitz_constants(Xs, kernelize="XTX", random_state=None):
    """Compute Lipschitz constants of gradients of linear regression problems.

    Find the largest eigenvalue of X^TX for several X, using power iteration.

    Parameters
    ----------
    Xs : array of shape (n_kernels, n_samples, n_features) or \
            (n_kernels, n_samples, n_samples)
        Multiple linear features or kernels.
    kernelize : str in {"XTX", "XXT", "X"}
        Whether to consider X^TX, XX^T, or directly X.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.

    Returns
    -------
    lipschitz : array of shape (n_kernels)
        Lipschitz constants.
    """
    backend = get_backend()

    if kernelize == "XXT":
        XTs = backend.transpose(Xs, (0, 2, 1))
        kernels = backend.matmul(Xs, XTs)
        del XTs
    elif kernelize == "XTX":
        XTs = backend.transpose(Xs, (0, 2, 1))
        kernels = backend.matmul(XTs, Xs)
        del XTs
    elif kernelize == "X":
        kernels = Xs
    else:
        raise ValueError("Unknown parameter kernelize=%r" % (kernelize, ))

    # check the random state
    random_generator = check_random_state(random_state)
    ys = random_generator.randn(*(kernels.shape[:2] + (1, )))

    ys = backend.asarray_like(ys, Xs)
    for i in range(10):
        ys /= backend.norm(ys, axis=1, keepdims=True) + 1e-16
        ys = backend.matmul(kernels, ys)
    evs = backend.norm(ys, axis=1)[:, 0]
    return evs


def assert_array_almost_equal(x, y, decimal=6, err_msg='', verbose=True):
    """Test array equality, casting all arrays to numpy."""
    backend = get_backend()
    x = backend.to_numpy(x)
    y = backend.to_numpy(y)
    return np.testing.assert_array_almost_equal(x, y, decimal=decimal,
                                                err_msg=err_msg,
                                                verbose=verbose)



import pytest

import numpy as np
import sklearn.linear_model
import sklearn.model_selection
import scipy.linalg

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.scoring import r2_score
from himalaya.utils import assert_array_almost_equal

from himalaya.kernel_ridge._hyper_gradient import _compute_delta_gradient
from himalaya.kernel_ridge._hyper_gradient import _compute_delta_loss
from himalaya.kernel_ridge import solve_multiple_kernel_ridge_hyper_gradient
from himalaya.kernel_ridge import solve_multiple_kernel_ridge_random_search
from himalaya.kernel_ridge import solve_weighted_kernel_ridge_conjugate_gradient  # noqa
from himalaya.kernel_ridge import generate_dirichlet_samples
from himalaya.kernel_ridge import predict_and_score_weighted_kernel_ridge


def _create_dataset(backend):
    n_featuress = [100, 200, 150]
    n_samples_train = 80
    n_samples_val = 20
    n_targets = 4

    Xs = [
        backend.asarray(backend.randn(n_samples_train, n_features),
                        backend.float64) for n_features in n_featuress
    ]
    Ks = backend.stack([X @ X.T for X in Xs])
    Xs_val = [
        backend.asarray(backend.randn(n_samples_val, n_features),
                        backend.float64) for n_features in n_featuress
    ]
    Ks_val = backend.stack([X_val @ X.T for X, X_val in zip(Xs, Xs_val)])

    true_gammas = backend.asarray(backend.rand(len(Xs), n_targets),
                                  backend.float64)

    ws = [
        backend.asarray(backend.randn(n_features, n_targets), backend.float64)
        for n_features in n_featuress
    ]
    Ys = backend.stack(
        [X @ w * backend.sqrt(g) for X, w, g in zip(Xs, ws, true_gammas)])
    Y = Ys.sum(0)
    Ys_val = backend.stack(
        [X @ w * backend.sqrt(g) for X, w, g in zip(Xs_val, ws, true_gammas)])
    Y_val = Ys_val.sum(0)

    gammas = backend.asarray(backend.rand(len(Xs), n_targets), backend.float64)

    dual_weights = backend.asarray(backend.randn(*Y.shape), backend.float64)

    return Ks, Y, dual_weights, gammas, Ks_val, Y_val, Xs


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_delta_gradient_direct(backend, n_targets_batch):
    backend = set_backend(backend)

    Ks, Y, dual_weights, gammas, Ks_val, Y_val, _ = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-1, 1, Y.shape[1]), Ks)
    deltas = backend.log(gammas / alphas)
    epsilons = backend.asarray_like(backend.randn(*deltas.shape), Ks)
    epsilons /= backend.norm(epsilons, axis=1)[:, None]
    step = 0.0000001
    deltas2 = deltas + epsilons * step

    # check direct gradient with a finite difference
    gradients = _compute_delta_gradient(Ks, Y, deltas, dual_weights,
                                        hyper_gradient_method='direct')[0]
    scores = _compute_delta_loss(Ks, Y, deltas, dual_weights)
    scores2 = _compute_delta_loss(Ks, Y, deltas2, dual_weights)

    directional_derivatives = (scores2 - scores) / step
    gradient_direction_product = (gradients * epsilons[:, :]).sum(0)
    norm = backend.norm(gradient_direction_product)

    assert_array_almost_equal(gradient_direction_product / norm,
                              directional_derivatives / norm, decimal=5)


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_delta_gradient_indirect(backend, n_targets_batch):
    backend = set_backend(backend)

    Ks, Y, _, gammas, Ks_val, Y_val, _ = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-1, 1, Y.shape[1]), Ks)
    deltas = backend.log(gammas / alphas)
    epsilons = backend.asarray_like(backend.randn(*deltas.shape), Ks)
    epsilons /= backend.norm(epsilons, axis=1)[:, None]
    step = 0.0000001
    deltas2 = deltas + epsilons * step

    # check direct and indirect gradient with a finite difference
    # to get the indirect gradient, we need to refit the kernel ridge during
    # the validation loss computation.

    def score_func(Y_val, Ypred):
        return 0.5 * backend.norm(Ypred - Y_val, axis=0) ** 2

    def compute_loss(deltas):
        dual_weights = solve_weighted_kernel_ridge_conjugate_gradient(
            Ks, Y, deltas=deltas, initial_dual_weights=None, alpha=1,
            max_iter=1000, tol=1e-5)
        loss = predict_and_score_weighted_kernel_ridge(Ks_val, dual_weights,
                                                       deltas, Y_val,
                                                       score_func=score_func)
        return loss, dual_weights

    loss, dual_weights = compute_loss(deltas)
    loss2, dual_weights2 = compute_loss(deltas2)

    gradients = _compute_delta_gradient(
        Ks_val, Y_val, deltas, dual_weights, Ks_train=Ks,
        hyper_gradient_method='conjugate_gradient', tol=1e-5)[0]

    directional_derivatives = (loss2 - loss) / step
    gradient_direction_product = (gradients * epsilons[:, :]).sum(0)
    norm = backend.norm(gradient_direction_product)

    assert_array_almost_equal(gradient_direction_product / norm,
                              directional_derivatives / norm, decimal=4)


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_hyper_gradient_n_targets_batch(
        backend, n_targets_batch):
    _test_solve_multiple_kernel_ridge_hyper_gradient(
        backend=backend, n_targets_batch=n_targets_batch)


@pytest.mark.parametrize('method', ["direct", "conjugate_gradient", "neumann"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_hyper_gradient_method(backend, method):
    _test_solve_multiple_kernel_ridge_hyper_gradient(backend=backend,
                                                     method=method)


@pytest.mark.parametrize('initial_deltas', [0, 5, 'ridgecv'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_hyper_gradient_initial_deltas(
        backend, initial_deltas):
    _test_solve_multiple_kernel_ridge_hyper_gradient(
        backend=backend, initial_deltas=initial_deltas)


@pytest.mark.parametrize('kernel_ridge',
                         ["conjugate_gradient", "gradient_descent"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_hyper_gradient_kernel_ridge(
        backend, kernel_ridge):
    _test_solve_multiple_kernel_ridge_hyper_gradient(backend=backend,
                                                     kernel_ridge=kernel_ridge)


def _test_solve_multiple_kernel_ridge_hyper_gradient(
        backend, n_targets_batch=None, method="direct", initial_deltas=0,
        kernel_ridge="conjugate_gradient"):
    backend = set_backend(backend)
    Ks, Y, dual_weights, gammas, Ks_val, Y_val, Xs = _create_dataset(backend)
    cv = 3
    progress_bar = False

    # compare bilinear gradient descent and dirichlet sampling
    alphas = backend.logspace(-5, 5, 11)
    gammas = generate_dirichlet_samples(50, len(Ks), concentration=[.1, 1.],
                                        random_state=0)
    _, _, cv_scores = \
        solve_multiple_kernel_ridge_random_search(
            Ks, Y, gammas, alphas, n_targets_batch=n_targets_batch,
            score_func=r2_score, cv=cv, progress_bar=progress_bar)
    scores_2 = backend.max(backend.asarray(cv_scores), axis=0)

    max_iter = 10
    for _ in range(3):
        try:
            _, _, cv_scores = \
                solve_multiple_kernel_ridge_hyper_gradient(
                    Ks, Y, max_iter=max_iter, n_targets_batch=n_targets_batch,
                    max_iter_inner_dual=1, max_iter_inner_hyper=1, tol=None,
                    score_func=r2_score, cv=cv,
                    hyper_gradient_method=method,
                    initial_deltas=initial_deltas,
                    kernel_ridge_method=kernel_ridge,
                    progress_bar=progress_bar)
            cv_scores = backend.asarray(cv_scores)
            scores_1 = cv_scores[cv_scores.sum(axis=1) != 0][-1]

            assert_array_almost_equal(scores_1, scores_2, decimal=1)
            break
        except AssertionError:
            max_iter *= 5
    else:
        raise AssertionError


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('return_weights', ['primal', 'dual'])
@pytest.mark.parametrize('method', ['hyper_gradient', 'random_search'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_return_weights(backend, method,
                                                    return_weights,
                                                    n_targets_batch):
    backend = set_backend(backend)

    Ks, Y, _, _, Ks_val, Y_val, Xs = _create_dataset(backend)
    n_targets = Y.shape[1]
    cv = sklearn.model_selection.check_cv(10)

    ############
    # run solver
    if method == "hyper_gradient":
        results = solve_multiple_kernel_ridge_hyper_gradient(
            Ks, Y, score_func=r2_score, cv=cv, max_iter=1,
            n_targets_batch=n_targets_batch, Xs=Xs, progress_bar=False,
            return_weights=return_weights)
        best_deltas, refit_weights, cv_scores = results
    elif method == "random_search":
        alphas = backend.asarray_like(backend.logspace(-3, 5, 2), Ks)
        results = solve_multiple_kernel_ridge_random_search(
            Ks, Y, n_iter=1, alphas=alphas, score_func=r2_score, cv=cv,
            n_targets_batch=n_targets_batch, Xs=Xs, progress_bar=False,
            return_weights=return_weights)
        best_deltas, refit_weights, cv_scores = results
    else:
        raise ValueError("Unknown parameter method=%r." % (method, ))

    ######################
    # test refited_weights
    for tt in range(n_targets):
        gamma = backend.exp(best_deltas[:, tt])
        alpha = 1.0

        if return_weights == 'primal':
            # compare primal weights with sklearn.linear_model.Ridge
            X = backend.concatenate(
                [t * backend.sqrt(g) for t, g in zip(Xs, gamma)], 1)
            model = sklearn.linear_model.Ridge(fit_intercept=False,
                                               alpha=backend.to_numpy(alpha))
            w1 = model.fit(backend.to_numpy(X),
                           backend.to_numpy(Y[:, tt])).coef_
            w1 = np.split(w1, np.cumsum([X.shape[1] for X in Xs][:-1]), axis=0)
            w1 = [backend.asarray(w) for w in w1]
            w1_scaled = backend.concatenate(
                [w * backend.sqrt(g) for w, g, in zip(w1, gamma)])
            assert_array_almost_equal(w1_scaled, refit_weights[:, tt],
                                      decimal=5)

        elif return_weights == 'dual':
            # compare dual weights with scipy.linalg.solve
            Ks_64 = backend.asarray(Ks, dtype=backend.float64)
            gamma_64 = backend.asarray(gamma, dtype=backend.float64)
            K = backend.matmul(Ks_64.T, gamma_64).T
            reg = backend.asarray_like(np.eye(K.shape[0]), K) * alpha
            Y_64 = backend.asarray(Y, dtype=backend.float64)
            c1 = scipy.linalg.solve(backend.to_numpy(K + reg),
                                    backend.to_numpy(Y_64[:, tt]))
            c1 = backend.asarray_like(c1, K)
            assert_array_almost_equal(c1, refit_weights[:, tt], decimal=5)



"""
Fitting a model on GPU
======================

This example demonstrates how to fit a model using GPU computations.

Himalaya implements different computational backends to fit the models:

- "numpy" (CPU) (default)
- "torch" (CPU)
- "torch_cuda" (GPU)
- "cupy" (GPU)

Each backend is only available if you installed the corresponding package with
CUDA enabled. Check the ``pytorch``/``cupy`` documentation for installation
instructions.
"""

###############################################################################
# Create a random dataset
# -----------------------
import numpy as np
n_samples, n_features, n_targets = 10, 5, 4
X = np.random.randn(n_samples, n_features)
Y = np.random.randn(n_samples, n_targets)

###############################################################################
# Change backend
# --------------
# To change the backend, you need to call the function
# ``himalaya.backend.set_backend``. With the option ``on_error="warn"``, the
# function does not raise an error if the new backend fails to be imported, and
# the backend is kept unchanged.

from himalaya.backend import set_backend
backend = set_backend("cupy", on_error="warn")

###############################################################################
# GPU backend
# -----------
# To fit a himalaya model on GPU, you don't need to move the input arrays to
# GPU, the method ``fit`` will do it for you. However, the float precision will
# not be changed.
#
# To make the most of GPU memory and computational speed, you might want to
# change the float precision to float32.
X = X.astype("float32")

from himalaya.kernel_ridge import KernelRidge
model_him = KernelRidge(kernel="linear", alpha=0.1)
model_him.fit(X, Y)

###############################################################################
# The results are stored in GPU memory, using an array object specific to the
# backend used. To use the results in other libraries (for example matplotlib),
# you can create a numpy array using the function ``backend.to_numpy``.
scores = model_him.score(X, Y)
print(scores.__class__)
scores = backend.to_numpy(scores)
print(scores.__class__)



import numbers

from ..backend import get_backend
from ..utils import compute_lipschitz_constants
from ._kernels import KernelCenterer


def _weighted_kernel_ridge_gradient(Ks, Y, dual_weights, exp_deltas, alpha=1.,
                                    double_K=False, return_objective=False):
    """Compute gradient of dual weights over a multi-kernel ridge regression.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Input kernels.
    Y : array of shape (n_samples, n_targets)
        Target data.
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    exp_deltas : array of shape (n_kernels, ) or (n_kernels, n_targets)
        Kernel weights.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    double_K : bool
        If True, multiply the gradient by the kernel to obtain the true
        gradients, which are less well conditionned.
    return_objective : bool
        If True, returns the objective function as well as the gradient.

    Returns
    -------
    dual_weight_gradient : array of shape (n_samples, n_targets)
        Gradients of `dual_weights`.
    objective : array of shape (n_targets, )
        Objective function.
    """
    backend = get_backend()

    if exp_deltas.ndim == 1:
        exp_deltas = exp_deltas[:, None]

    Ks_times_dualweights = backend.stack(
        [backend.matmul(K, dual_weights).T for K in Ks], 2)

    predictions = backend.matmul(Ks_times_dualweights,
                                 exp_deltas.T[..., None])[..., 0].T
    residual = predictions - Y
    dampened_residual = residual + dual_weights * alpha

    if double_K:
        Ks_times_dampened_residual = backend.stack(
            [backend.matmul(K, dampened_residual).T for K in Ks], 2)
        dual_weight_gradient = backend.matmul(
            Ks_times_dampened_residual, exp_deltas.T[..., None])[..., 0].T
    else:
        dual_weight_gradient = dampened_residual

    if return_objective:
        objective = backend.norm(residual, axis=0) ** 2
        objective += alpha * (predictions * dual_weights).sum(0)

        return dual_weight_gradient, objective
    else:
        return dual_weight_gradient


def solve_weighted_kernel_ridge_gradient_descent(
        Ks, Y, deltas, alpha=1., fit_intercept=False, step_sizes=None,
        lipschitz_Ks=None, initial_dual_weights=None, max_iter=100, tol=1e-3,
        double_K=False, random_state=None, debug=False):
    """Solve weighted kernel ridge regression using gradient descent.

    Solve the kernel ridge regression::

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    where the kernel K is a weighted sum of multiple kernels::

        K = sum_i exp(deltas[i]) Ks[i]

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Input kernels.
    Y : array of shape (n_samples, n_targets)
        Target data.
    deltas : array of shape (n_kernels, ) or (n_kernels, n_targets)
        Kernel weights.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept. If False, Ks should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    step_sizes : float, or array of shape (n_targets), or None
        Step sizes.
        If None, computes a step size based on the Lipschitz constants.
    lipschitz_Ks : float, or array of shape (n_kernels), or None:
        Lipschitz constant.
        Used only if `step_sizes` is None.
        If None, Lipschitz constants are estimated with power iteration on Ks.
    initial_dual_weights : array of shape (n_samples, n_targets)
        Initial kernel ridge coefficients.
    max_iter : int
        Maximum number of gradient step.
    tol : float > 0 or None
        Tolerance for the stopping criterion.
    double_K : bool
        If True, multiply the gradient by the kernel to obtain the true
        gradients, which are less well conditionned.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.
    debug : bool
        If True, check some intermediate computations.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    n_targets = Y.shape[1]

    if deltas.ndim == 1:
        deltas = deltas[:, None]
    if isinstance(alpha, numbers.Number) or alpha.ndim == 0:
        alpha = backend.ones_like(Y, shape=(1, )) * alpha

    Ks, Y, deltas, alpha, step_sizes, lipschitz_Ks, initial_dual_weights = \
        backend.check_arrays(Ks, Y, deltas, alpha, step_sizes, lipschitz_Ks,
                             initial_dual_weights)
    exp_deltas = backend.exp(deltas)

    if fit_intercept:
        Ks, Y, Ks_rows, Y_offset = _helper_intercept(Ks, Y)

    if initial_dual_weights is None:
        dual_weights = backend.zeros_like(Y)
    else:
        dual_weights = backend.copy(initial_dual_weights)

    if step_sizes is None:
        if lipschitz_Ks is None:
            lipschitz_Ks = compute_lipschitz_constants(
                Ks, random_state=random_state)
        if not double_K:
            lipschitz_Ks = backend.sqrt(lipschitz_Ks)

        total_lip = backend.matmul(lipschitz_Ks[None, :],
                                   exp_deltas)[0] + alpha
        step_sizes = 1. / total_lip

        if debug:
            assert not backend.any(backend.isnan(step_sizes))

    if isinstance(step_sizes, numbers.Number) or step_sizes.ndim == 0:
        step_sizes = backend.ones_like(Y, shape=(1, )) * step_sizes

    #######################
    # Gradient descent loop
    converged = backend.zeros_like(Y, dtype=backend.bool, shape=(n_targets))
    for i in range(max_iter):
        grads = _weighted_kernel_ridge_gradient(Ks, Y[:, ~converged],
                                                dual_weights[:, ~converged],
                                                exp_deltas=exp_deltas,
                                                alpha=alpha, double_K=double_K)
        update = step_sizes * grads
        dual_weights[:, ~converged] -= update

        ##########################
        # remove converged targets
        if tol is not None:
            relative_update = backend.abs(update / dual_weights[:, ~converged])
            relative_update[dual_weights[:, ~converged] == 0] = 0
            just_converged = backend.max(relative_update, 0) < tol

            if exp_deltas.shape[1] == just_converged.shape[0]:
                exp_deltas = exp_deltas[:, ~just_converged]
            if alpha.shape[0] == just_converged.shape[0]:
                alpha = alpha[~just_converged]
            if step_sizes.shape[0] == just_converged.shape[0]:
                step_sizes = step_sizes[~just_converged]
            converged[~converged] = just_converged

            if backend.all(converged):
                break

    if fit_intercept:
        intercept = Y_offset
        intercept -= ((Ks_rows @ dual_weights) * backend.exp(deltas)).sum(0)
        return dual_weights, intercept
    else:
        return dual_weights


def solve_weighted_kernel_ridge_conjugate_gradient(Ks, Y, deltas, alpha=1.,
                                                   fit_intercept=False,
                                                   initial_dual_weights=None,
                                                   max_iter=100, tol=1e-4,
                                                   random_state=None):
    """Solve weighted kernel ridge regression using conjugate gradient.

    Solve the kernel ridge regression::

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    where the kernel K is a weighted sum of multiple kernels::

        K = sum_i exp(deltas[i]) Ks[i]

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Input kernels.
    Y : torch.Tensor of shape (n_samples, n_targets)
        Target data.
    deltas : array of shape (n_kernels, ) or (n_kernels, n_targets)
        Kernel weights.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept. If False, Ks should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    initial_dual_weights : array of shape (n_samples, n_targets)
        Initial kernel ridge coefficients.
    max_iter : int
        Maximum number of conjugate gradient step.
    tol : float > 0 or None
        Tolerance for the stopping criterion.
    random_state : int, or None
        Random generator seed. Not used.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    n_targets = Y.shape[1]

    if deltas.ndim == 1:
        deltas = deltas[:, None]
    if isinstance(alpha, numbers.Number) or alpha.ndim == 0:
        alpha = backend.ones_like(Y, shape=(1, )) * alpha

    Ks, Y, deltas, alpha, initial_dual_weights = backend.check_arrays(
        Ks, Y, deltas, alpha, initial_dual_weights)
    exp_deltas = backend.exp(deltas)

    if fit_intercept:
        Ks, Y, Ks_rows, Y_offset = _helper_intercept(Ks, Y)

    if initial_dual_weights is None:
        dual_weights = backend.zeros_like(Y)
    else:
        dual_weights = backend.copy(initial_dual_weights)

    # compute initial residual
    r = _weighted_kernel_ridge_gradient(Ks, Y, dual_weights,
                                        exp_deltas=exp_deltas, alpha=alpha,
                                        double_K=False)
    r *= -1
    p = backend.copy(r)
    new_squared_residual_norm = backend.norm(r, axis=0) ** 2

    #########################
    # Conjugate gradient loop
    converged = backend.zeros_like(Y, dtype=backend.bool, shape=(n_targets))
    for i in range(max_iter):
        Ks_x_p = backend.matmul(Ks, p)
        tmp = backend.matmul(backend.transpose(Ks_x_p, (2, 1, 0)),
                             backend.transpose(exp_deltas,
                                               (1, 0))[:, :, None])[..., 0]
        K_x_p_plus_reg = backend.transpose(tmp, (1, 0)) + alpha * p

        squared_residual_norm = new_squared_residual_norm
        squared_p_A_norm = backend.matmul(
            backend.transpose(p, (1, 0))[:, None],
            backend.transpose(K_x_p_plus_reg, (1, 0))[..., None])[:, 0, 0]

        squared_p_A_norm[squared_p_A_norm == 0] = 1
        alpha_step = squared_residual_norm / squared_p_A_norm

        update = alpha_step * p
        dual_weights[:, ~converged] += update

        r -= alpha_step * K_x_p_plus_reg

        new_squared_residual_norm = backend.norm(r, axis=0) ** 2

        beta = (new_squared_residual_norm) / (squared_residual_norm)
        p = r + beta * p

        ##########################
        # remove converged targets
        if tol is not None:
            relative_update = backend.abs(update / dual_weights[:, ~converged])
            relative_update[dual_weights[:, ~converged] == 0] = 0
            just_converged = backend.max(relative_update, 0) < tol

            r = r[:, ~just_converged]
            p = p[:, ~just_converged]
            new_squared_residual_norm = \
                new_squared_residual_norm[~just_converged]
            if exp_deltas.shape[1] == just_converged.shape[0]:
                exp_deltas = exp_deltas[:, ~just_converged]
            if alpha.shape[0] == just_converged.shape[0]:
                alpha = alpha[~just_converged]
            converged[~converged] = just_converged
            if backend.all(converged):
                break

    if fit_intercept:
        intercept = Y_offset
        intercept -= ((Ks_rows @ dual_weights) * backend.exp(deltas)).sum(0)
        return dual_weights, intercept
    else:
        return dual_weights


def solve_weighted_kernel_ridge_neumann_series(Ks, Y, deltas, alpha=1.,
                                               fit_intercept=False,
                                               max_iter=10, factor=0.0001,
                                               tol=None, random_state=None,
                                               debug=False):
    """Solve weighted kernel ridge regression using Neumann series.

    Solve the kernel ridge regression::

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    where the kernel K is a weighted sum of multiple kernels::

        K = sum_i exp(deltas[i]) Ks[i]

    The Neumann series approximate the invert of K as K^-1 = sum_j (Id - K)^j.
    It is a poor approximation, so this solver should NOT be used to solve
    ridge. It is however useful during hyper-parameter gradient descent, as
    we do not need a good precision of the results, but merely the direction
    of the gradient.

    See [Lorraine, Vicol, & Duvenaud (2019). Optimizing Millions of
    Hyperparameters by Implicit Differentiation. arXiv:1911.02590].

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Input kernels.
    Y : torch.Tensor of shape (n_samples, n_targets)
        Target data.
    deltas : array of shape (n_kernels, ) or (n_kernels, n_targets)
        Kernel weights.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept. If False, Ks should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    max_iter : int
        Number of terms in the Neumann series.
    factor : float, or array of shape (n_targets, )
        Factor used to allow convergence of the series. We actually invert
        (factor * K) instead of K, then multiply the result by factor.
    tol : None
        Not used.
    random_state : int, or None
        Random generator seed. Not used.
    debug : bool
        If True, check some intermediate computations.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()

    if deltas.ndim == 1:
        deltas = deltas[:, None]
    if isinstance(alpha, numbers.Number) or alpha.ndim == 0:
        alpha = backend.ones_like(Y, shape=(1, )) * alpha
    if isinstance(factor, numbers.Number) or factor.ndim == 0:
        factor = backend.ones_like(Y, shape=(1, )) * factor

    Ks, Y, deltas, alpha, factor = backend.check_arrays(
        Ks, Y, deltas, alpha, factor)
    exp_deltas = backend.exp(deltas)

    if fit_intercept:
        Ks, Y, Ks_rows, Y_offset = _helper_intercept(Ks, Y)

    # product accumulator: product = (id_minus_K ** ii) @ Ys
    product = Y
    # sum accumulator: dual_weights = sum_ii product
    dual_weights = backend.zeros_like(Y)
    for ii in range(max_iter):
        product = (
            product * (1 - factor[None, :] * alpha[None, :]) -
            factor[None, :] * backend.sum(
                exp_deltas[:, None, :] * backend.matmul(Ks, product), axis=0))
        dual_weights += product

    dual_weights *= factor[None, :]

    if debug:
        assert not backend.any(backend.isinf(dual_weights))
        assert not backend.any(backend.isnan(dual_weights))

    if fit_intercept:
        intercept = Y_offset
        intercept -= ((Ks_rows @ dual_weights) * backend.exp(deltas)).sum(0)
        return dual_weights, intercept
    else:
        return dual_weights


def _helper_intercept(Ks, Y):
    """Transform Ks and Y if we fit an intercept."""
    backend = get_backend()

    # unfortunate, but we need to preserve the input. XXX add a copy parameter
    Ks = backend.copy(Ks)
    centerer = KernelCenterer()
    Ks_rows = backend.zeros_like(Ks, shape=(Ks.shape[0], Ks.shape[1]))
    for ii in range(Ks.shape[0]):
        Ks[ii] = centerer.fit_transform(Ks[ii])
        Ks_rows[ii] = centerer.K_fit_rows_

    Y = backend.copy(Y)
    Y_offset = Y.mean(0)
    Y -= Y_offset
    return Ks, Y, Ks_rows, Y_offset


#: Dictionary with all weighted kernel ridge solvers.
WEIGHTED_KERNEL_RIDGE_SOLVERS = {
    "neumann_series": solve_weighted_kernel_ridge_neumann_series,
    "conjugate_gradient": solve_weighted_kernel_ridge_conjugate_gradient,
    "gradient_descent": solve_weighted_kernel_ridge_gradient_descent,
}
###############################################################################


def solve_kernel_ridge_conjugate_gradient(K, Y, alpha=1., fit_intercept=False,
                                          initial_dual_weights=None,
                                          max_iter=100, tol=1e-3,
                                          random_state=None):
    """Solve kernel ridge regression using conjugate gradient.

    Solve the kernel ridge regression::

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    Parameters
    ----------
    K : array of shape (n_samples, n_samples)
        Input kernel.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept. If False, K should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    initial_dual_weights : array of shape (n_samples, n_targets)
        Initial kernel ridge coefficients.
    max_iter : int
        Maximum number of conjugate gradient step.
    tol : float > 0 or None
        Tolerance for the stopping criterion.
    random_state : int, or None
        Random generator seed. Not used.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    deltas = backend.zeros_like(K, shape=(1, ))
    return solve_weighted_kernel_ridge_conjugate_gradient(
        K[None], Y=Y, deltas=deltas, alpha=alpha, fit_intercept=fit_intercept,
        initial_dual_weights=initial_dual_weights, max_iter=max_iter, tol=tol)


def solve_kernel_ridge_gradient_descent(K, Y, alpha=1., fit_intercept=False,
                                        step_sizes=None, lipschitz_Ks=None,
                                        initial_dual_weights=None,
                                        max_iter=100, tol=1e-3, double_K=False,
                                        random_state=None, debug=False):
    """Solve kernel ridge regression using conjugate gradient.

    Solve the kernel ridge regression

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    Parameters
    ----------
    K : array of shape (n_samples, n_samples)
        Input kernel.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept. If False, K should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    step_sizes : float, or array of shape (n_targets), or None
        Step sizes.
        If None, computes a step size based on the Lipschitz constants.
    lipschitz_Ks : float, or array of shape (n_kernels), or None:
        Lipschitz constant.
        Used only if `step_sizes` is None.
        If None, Lipschitz constants are estimated with power iteration on Ks.
    initial_dual_weights : array of shape (n_samples, n_targets)
        Initial kernel ridge coefficients.
    max_iter : int
        Maximum number of gradient step.
    tol : float > 0 or None
        Tolerance for the stopping criterion.
    double_K : bool
        If True, multiply the gradient by the kernel to obtain the true
        gradients, which are less well conditionned.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.
    debug : bool
        If True, check some intermediate computations.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    deltas = backend.zeros_like(K, shape=(1, ))
    return solve_weighted_kernel_ridge_gradient_descent(
        K[None], Y=Y, deltas=deltas, alpha=alpha, step_sizes=step_sizes,
        lipschitz_Ks=lipschitz_Ks, initial_dual_weights=initial_dual_weights,
        max_iter=max_iter, tol=tol, double_K=double_K,
        random_state=random_state, debug=debug, fit_intercept=fit_intercept)


def solve_kernel_ridge_eigenvalues(K, Y, alpha=1., method="eigh",
                                   fit_intercept=False,
                                   negative_eigenvalues="zeros",
                                   random_state=None):
    """Solve kernel ridge regression using eigenvalues decomposition.

    Solve the kernel ridge regression::

        w* = argmin_w ||K @ w - Y||^2 + alpha (w.T @ K @ w)

    Parameters
    ----------
    K : array of shape (n_samples, n_samples)
        Input kernel.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    method : str in {"eigh", "svd"}
        Method used to diagonalize the kernel.
    fit_intercept : boolean
        Whether to fit an intercept. If False, K should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    negative_eigenvalues : str in {"nan", "error", "zeros"}
        If the decomposition leads to negative eigenvalues (wrongly emerging
        from float32 errors):
        - "error" raises an error.
        - "zeros" remplaces them with zeros.
        - "nan" returns nans if the regularization does not compensate
        twice the smallest negative value, else it ignores the problem.
    random_state : int, or None
        Random generator seed. Not used.

    Returns
    -------
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    if isinstance(alpha, numbers.Number) or alpha.ndim == 0:
        alpha = backend.ones_like(Y, shape=(1, )) * alpha

    K, Y, alpha = backend.check_arrays(K, Y, alpha)

    centerer, Y_offset = None, None
    if fit_intercept:
        centerer = KernelCenterer()
        K = centerer.fit_transform(K)
        Y_offset = Y.mean(0)
        Y = Y - Y_offset

    if method == "eigh":
        # diagonalization: K = V @ np.diag(eigenvalues) @ V.T
        eigenvalues, V = backend.eigh(K)
        # match SVD notations: K = U @ np.diag(eigenvalues) @ Vt
        U = V
        Vt = V.T
    elif method == "svd":
        # SVD: K = U @ np.diag(eigenvalues) @ Vt
        U, eigenvalues, Vt = backend.svd(K)
    else:
        raise ValueError("Unknown method=%r." % (method, ))

    inverse = 1 / (alpha[None] + eigenvalues[:, None])

    # negative eigenvalues can emerge from incorrect kernels, or from float32
    if eigenvalues[0] < 0:
        if negative_eigenvalues == "nan":
            if alpha < -eigenvalues[0] * 2:
                return backend.ones_like(Y) * backend.asarray(
                    backend.nan, dtype=Y.dtype)
            else:
                pass

        elif negative_eigenvalues == "zeros":
            eigenvalues[eigenvalues < 0] = 0

        elif negative_eigenvalues == "error":
            raise RuntimeError(
                "Negative eigenvalues. Make sure the kernel is positive "
                "semi-definite, increase the regularization alpha, or use"
                "another solver.")
        else:
            raise ValueError("Unknown negative_eigenvalues=%r." %
                             (negative_eigenvalues, ))

    iUT = inverse[:, None, :] * U.T[:, :, None]
    if Y.shape[0] < Y.shape[1]:
        dual_weights = (
            backend.transpose(Vt.T @ iUT,
                              (2, 0, 1)) @ Y.T[:, :, None])[:, :, 0].T
    else:
        dual_weights = Vt.T @ (
            backend.transpose(iUT, (2, 0, 1)) @ Y.T[:, :, None])[:, :, 0].T

    if fit_intercept:
        intercept = Y_offset - centerer.K_fit_rows_ @ dual_weights
        return dual_weights, intercept
    else:
        return dual_weights


#: Dictionary with all kernel ridge solvers
KERNEL_RIDGE_SOLVERS = {
    "eigenvalues": solve_kernel_ridge_eigenvalues,
    "conjugate_gradient": solve_kernel_ridge_conjugate_gradient,
    "gradient_descent": solve_kernel_ridge_gradient_descent,
}



import numpy as np


def plot_alphas_diagnostic(best_alphas, alphas, ax=None):
    """Plot a diagnostic plot for the selected alphas during cross-validation.

    To figure out whether to increase the range of alphas.

    Parameters
    ----------
    best_alphas : array of shape (n_targets, )
        Alphas selected during cross-validation for each target.
    alphas : array of shape (n_alphas)
        Alphas used while fitting the model.
    ax : None or figure axis

    Returns
    -------
    ax : figure axis
    """
    import matplotlib.pyplot as plt
    alphas = np.sort(alphas)
    n_alphas = len(alphas)
    indices = np.argmin(np.abs(best_alphas[None] - alphas[:, None]), 0)
    hist = np.bincount(indices, minlength=n_alphas)

    if ax is None:
        fig, ax = plt.subplots(1, 1)

    log10alphas = np.log(alphas) / np.log(10)
    ax.plot(log10alphas, hist, '.-', markersize=12)
    ax.set_ylabel('Number of targets')
    ax.set_xlabel('log10(alpha)')
    ax.grid("on")
    return ax



"""
Multiple kernel ridge regression
================================
This example demonstrates how to solve multiple kernel ridge regression.
It uses random search and cross validation to select optimal hyperparameters.
"""

import numpy as np
import matplotlib.pyplot as plt

from himalaya.backend import set_backend
from himalaya.kernel_ridge import solve_multiple_kernel_ridge_random_search
from himalaya.kernel_ridge import predict_and_score_weighted_kernel_ridge
from himalaya.scoring import r2_score_split
from himalaya.viz import plot_alphas_diagnostic

# sphinx_gallery_thumbnail_number = 4
###############################################################################
# In this example, we use the ``cupy`` backend, and fit the model on GPU.

backend = set_backend("cupy")

###############################################################################
# Generate a random dataset
# -------------------------
#
# - Xs_train : list of arrays of shape (n_samples_train, n_features)
# - Xs_test : list of arrays of shape (n_samples_test, n_features)
# - Y_train : array of shape (n_samples_train, n_targets)
# - Y_test : array of shape (n_repeat, n_samples_test, n_targets)

n_samples_train = 1000
n_samples_test = 300
n_targets = 1000
n_features_list = [1000, 1000, 500]

Xs_train = [
    backend.randn(n_samples_train, n_features)
    for n_features in n_features_list
]
Xs_test = [
    backend.randn(n_samples_test, n_features) for n_features in n_features_list
]
ws = [
    backend.randn(n_features, n_targets) / n_features
    for n_features in n_features_list
]
Y_train = backend.stack([X @ w for X, w in zip(Xs_train, ws)]).sum(0)
Y_test = backend.stack([X @ w for X, w in zip(Xs_test, ws)]).sum(0)

###############################################################################
# Optional: Add some arbitrary scalings per kernel
if True:
    scalings = [0.2, 5, 1]
    Xs_train = [X * scaling for X, scaling in zip(Xs_train, scalings)]
    Xs_test = [X * scaling for X, scaling in zip(Xs_test, scalings)]

Y_train -= Y_train.mean(0)
Y_test -= Y_test.mean(0)

###############################################################################
# Precompute the linear kernels
# -----------------------------
# We also cast them to float32.

Ks_train = backend.stack([X_train @ X_train.T for X_train in Xs_train])
Ks_train = backend.asarray(Ks_train, dtype=backend.float32)
Y_train = backend.asarray(Y_train, dtype=backend.float32)

Ks_test = backend.stack(
    [X_test @ X_train.T for X_train, X_test in zip(Xs_train, Xs_test)])
Ks_test = backend.asarray(Ks_test, dtype=backend.float32)
Y_test = backend.asarray(Y_test, dtype=backend.float32)

###############################################################################
# Run the solver, using random search
# -----------------------------------
# This method should work fine for
# small number of kernels (< 20). The larger the number of kenels, the larger
# we need to sample the hyperparameter space (i.e. increasing ``n_iter``).

###############################################################################
# Here we use 100 iterations to have a reasonably fast example (~40 sec).
# To have a better convergence, we probably need more iterations.
# Note that there is currently no stopping criterion in this method.
n_iter = 100

###############################################################################
# Grid of regularization parameters.
alphas = np.logspace(-10, 10, 21)

###############################################################################
# Batch parameters are used to reduce the necessary GPU memory. A larger value
# will be a bit faster, but the solver might crash if it runs out of memory.
# Optimal values depend on the size of your dataset.
n_targets_batch = 1000
n_alphas_batch = 20

###############################################################################
# If ``return_weights == "dual"``, the solver will use more memory.
# To mitigate this, you can reduce ``n_targets_batch`` in the refit
# using ```n_targets_batch_refit``.
# If you don't need the dual weights, use ``return_weights = None``.
return_weights = 'dual'
n_targets_batch_refit = 200

###############################################################################
# Run the solver. For each iteration, it will:
#
# - sample kernel weights from a Dirichlet distribution
# - fit (n_splits * n_alphas * n_targets) ridge models
# - compute the scores on the validation set of each split
# - average the scores over splits
# - take the maximum over alphas
# - (only if you ask for the ridge weights) refit using the best alphas per
#   target and the entire dataset
# - return for each target the log kernel weights leading to the best CV score
#   (and the best weights if necessary)
results = solve_multiple_kernel_ridge_random_search(
    Ks=Ks_train,
    Y=Y_train,
    n_iter=n_iter,
    alphas=alphas,
    n_targets_batch=n_targets_batch,
    return_weights=return_weights,
    n_alphas_batch=n_alphas_batch,
    n_targets_batch_refit=n_targets_batch_refit,
    jitter_alphas=True,
)

###############################################################################
# As we used the ``cupy`` backend, the results are ``cupy`` arrays, which are
# on GPU. Here, we cast the results back to CPU, and to ``numpy`` arrays.
deltas = backend.to_numpy(results[0])
dual_weights = backend.to_numpy(results[1])
cv_scores = backend.to_numpy(results[2])

###############################################################################
# Plot the convergence curve
# --------------------------
#
# ``cv_scores`` gives the scores for each sampled kernel weights.
# The convergence curve is thus the current maximum for each target.

current_max = np.maximum.accumulate(cv_scores, axis=0)
mean_current_max = np.mean(current_max, axis=1)
x_array = np.arange(1, len(mean_current_max) + 1)
plt.plot(x_array, mean_current_max, '-o')
plt.grid("on")
plt.xlabel("Number of kernel weights sampled")
plt.ylabel("L2 negative loss (higher is better)")
plt.title("Convergence curve, averaged over targets")
plt.tight_layout()
plt.show()

###############################################################################
# Plot the optimal alphas selected by the solver
# ----------------------------------------------
#
# This plot is helpful to refine the alpha grid if the range is too small or
# too large.

best_alphas = 1. / np.sum(np.exp(deltas), axis=0)
plot_alphas_diagnostic(best_alphas, alphas)
plt.title("Best alphas selected by cross-validation")
plt.show()

###############################################################################
# Compute the predictions on the test set
# ---------------------------------------
# (requires the dual weights)

split = False
scores = predict_and_score_weighted_kernel_ridge(
    Ks_test, dual_weights, deltas, Y_test, split=split,
    n_targets_batch=n_targets_batch, score_func=r2_score_split)
scores = backend.to_numpy(scores)

plt.hist(scores, 50)
plt.xlabel(r"$R^2$ generalization score")
plt.title("Histogram over targets")
plt.show()

###############################################################################
# Compute the split predictions on the test set 
# ---------------------------------------------
# (requires the dual weights)
#
# Here we apply the dual weights on each kernel separately
# (``exp(deltas[i]) * kernel[i]``), and we compute the R\ :sup:`2` scores
# (corrected for correlations) of each prediction.

split = True
scores = predict_and_score_weighted_kernel_ridge(
    Ks_test, dual_weights, deltas, Y_test, split=split,
    n_targets_batch=n_targets_batch, score_func=r2_score_split)
scores = backend.to_numpy(scores)

bins = np.linspace(scores.min(), scores.max(), 50)
for score in scores:
    plt.hist(score, bins, alpha=0.5)
plt.title(r"Histogram of $R^2$ generalization score split between kernels")
plt.legend(["kernel %d" % kk for kk in range(scores.shape[0])])
plt.show()



import warnings
import numbers

import numpy as np
from sklearn.model_selection import check_cv

from ..backend import get_backend
from ..backend._utils import _dtype_to_str
from ..progress_bar import bar
from ..scoring import l2_neg_loss
from ..validation import check_random_state
from ..kernel_ridge import generate_dirichlet_samples
from ..kernel_ridge._random_search import _select_best_alphas


def solve_group_ridge_random_search(
        Xs, Y, n_iter=100, concentration=[0.1, 1.0
                                          ], alphas=1.0, fit_intercept=False,
        score_func=l2_neg_loss, cv=5, return_weights=False, local_alpha=True,
        jitter_alphas=False, random_state=None, n_targets_batch=None,
        n_targets_batch_refit=None, n_alphas_batch=None, progress_bar=True,
        conservative=False, Y_in_cpu=False, diagonalize_method="svd"):
    """Solve group ridge regression using random search on the simplex.

    Solve the group-regularized ridge regression::

        b* = argmin_b ||Z @ b - Y||^2 + ||b||^2

    where the feature space X_i is scaled by a group scaling ::

        Z_i = exp(deltas[i] / 2) X_i

    Parameters
    ----------
    Xs : list of len (n_spaces), with arrays of shape (n_samples, n_features)
        Input features.
    Y : array of shape (n_samples, n_targets)
        Target data.
    n_iter : int, or array of shape (n_iter, n_spaces)
        Number of feature-space weights combination to search.
        If an array is given, the solver uses it as the list of weights
        to try, instead of sampling from a Dirichlet distribution.
    concentration : float, or list of float
        Concentration parameters of the Dirichlet distribution.
        If a list, iteratively cycle through the list.
        Not used if n_iter is an array.
    alphas : float or array of shape (n_alphas, )
        Range of ridge regularization parameter. The log group-weights
        ``deltas`` are equal to log(gamma/alpha), where gamma is randomly
        sampled on the simplex, and alpha is selected from a list of
        candidates.
    fit_intercept : boolean
        Whether to fit an intercept.
        If False, Xs and Y must be zero-mean over samples.
    score_func : callable
        Function used to compute the score of predictions versus Y.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    return_weights : bool
        Whether to refit on the entire dataset and return the weights.
    local_alpha : bool
        If True, alphas are selected per target, else shared over all targets.
    jitter_alphas : bool
        If True, alphas range is slightly jittered for each gamma.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.
    n_targets_batch : int or None
        Size of the batch for over targets during cross-validation.
        Used for memory reasons. If None, uses all n_targets at once.
    n_targets_batch_refit : int or None
        Size of the batch for over targets during refit.
        Used for memory reasons. If None, uses all n_targets at once.
    n_alphas_batch : int or None
        Size of the batch for over alphas. Used for memory reasons.
        If None, uses all n_alphas at once.
    progress_bar : bool
        If True, display a progress bar over gammas.
    conservative : bool
        If True, when selecting the hyperparameter alpha, take the largest one
        that is less than one standard deviation away from the best.
        If False, take the best.
    Y_in_cpu : bool
        If True, keep the target values ``Y`` in CPU memory (slower).
    diagonalize_method : str in {"svd"}
        Method used to diagonalize the features.

    Returns
    -------
    deltas : array of shape (n_spaces, n_targets)
        Best log feature-space weights for each target.
    refit_weights : array of shape (n_features, n_targets), or None
        Refit regression weights on the entire dataset, using selected best
        hyperparameters. Refit weights are always stored on CPU memory.
    cv_scores : array of shape (n_iter, n_targets)
        Cross-validation scores per iteration, averaged over splits, for the
        best alpha. Cross-validation scores will always be on CPU memory.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    n_spaces = len(Xs)
    if isinstance(n_iter, int):
        gammas = generate_dirichlet_samples(n_samples=n_iter,
                                            n_kernels=n_spaces,
                                            concentration=concentration,
                                            random_state=random_state)
        gammas[0] = 1 / n_spaces
    elif n_iter.ndim == 2:
        gammas = n_iter
        assert gammas.shape[1] == n_spaces
    else:
        raise ValueError("Unknown parameter n_iter=%r." % (n_iter, ))

    if isinstance(alphas, numbers.Number) or alphas.ndim == 0:
        alphas = backend.ones_like(Y, shape=(1, )) * alphas

    dtype = Xs[0].dtype
    gammas = backend.asarray(gammas, dtype=dtype)
    device = getattr(gammas, "device", None)
    gammas, alphas = backend.check_arrays(gammas, alphas)
    Y = backend.asarray(Y, dtype=dtype, device="cpu" if Y_in_cpu else device)
    Xs = [backend.asarray(X, dtype=dtype, device=device) for X in Xs]

    # stack all features
    X_ = backend.concatenate(Xs, 1)
    n_features_list = [X.shape[1] for X in Xs]
    n_features = X_.shape[1]
    start_and_end = np.concatenate([[0], np.cumsum(n_features_list)])
    slices = [
        slice(start, end)
        for start, end in zip(start_and_end[:-1], start_and_end[1:])
    ]
    del Xs

    X_offset, Y_offset = None, None
    if fit_intercept:
        X_offset = X_.mean(0)
        Y_offset = Y.mean(0)
        X_ = X_ - X_offset
        Y = Y - Y_offset

    n_samples, n_targets = Y.shape
    if n_targets_batch is None:
        n_targets_batch = n_targets
    if n_targets_batch_refit is None:
        n_targets_batch_refit = n_targets_batch
    if n_alphas_batch is None:
        n_alphas_batch = len(alphas)

    cv = check_cv(cv)
    n_splits = cv.get_n_splits()
    for train, val in cv.split(Y):
        if len(val) == 0 or len(train) == 0:
            raise ValueError("Empty train or validation set. "
                             "Check that `cv` is correctly defined.")

    random_generator, given_alphas = None, None
    if jitter_alphas:
        random_generator = check_random_state(random_state)
        given_alphas = backend.copy(alphas)

    best_gammas = backend.full_like(gammas, fill_value=1.0 / n_spaces,
                                    shape=(n_spaces, n_targets))
    best_alphas = backend.ones_like(gammas, shape=n_targets)
    cv_scores = backend.zeros_like(gammas, shape=(len(gammas), n_targets),
                                   device="cpu")
    current_best_scores = backend.full_like(gammas, fill_value=-backend.inf,
                                            shape=n_targets)

    # initialize refit ridge weights
    refit_weights = None
    if return_weights:
        refit_weights = backend.zeros_like(gammas,
                                           shape=(n_features, n_targets),
                                           device="cpu")

    for ii, gamma in enumerate(
            bar(gammas, '%d random sampling with cv' % len(gammas),
                use_it=progress_bar)):

        for kk in range(n_spaces):
            X_[:, slices[kk]] *= backend.sqrt(gamma[kk])

        if jitter_alphas:
            noise = backend.asarray_like(random_generator.rand(), alphas)
            alphas = given_alphas * (10 ** (noise - 0.5))

        scores = backend.zeros_like(gammas,
                                    shape=(n_splits, len(alphas), n_targets))
        for jj, (train, test) in enumerate(cv.split(X_)):
            train = backend.to_gpu(train, device=device)
            test = backend.to_gpu(test, device=device)
            Xtrain, Xtest = X_[train], X_[test]

            if fit_intercept:
                Xtrain_mean = X_[train].mean(0)
                Xtrain = X_[train] - Xtrain_mean
                Xtest = X_[test] - Xtrain_mean

            for matrix, alpha_batch in _decompose_ridge(
                    Xtrain=Xtrain, alphas=alphas, negative_eigenvalues="nan",
                    n_alphas_batch=n_alphas_batch, method=diagonalize_method):
                # n_alphas_batch, n_features, n_samples_train = \
                # matrix.shape
                matrix = backend.matmul(Xtest, matrix)
                # n_alphas_batch, n_samples_test, n_samples_train = \
                # matrix.shape

                predictions = None
                for start in range(0, n_targets, n_targets_batch):
                    batch = slice(start, start + n_targets_batch)
                    Ytrain = backend.to_gpu(Y[:, batch][train], device=device)
                    Ytest = backend.to_gpu(Y[:, batch][test], device=device)
                    if fit_intercept:
                        Ytrain_mean = Ytrain.mean(0)
                        Ytrain = Ytrain - Ytrain_mean
                        Ytest = Ytest - Ytrain_mean

                    predictions = backend.matmul(matrix, Ytrain)
                    # n_alphas_batch, n_samples_test, n_targets_batch = \
                    # predictions.shape

                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore", category=UserWarning)
                        scores[jj, alpha_batch, batch] = score_func(
                            Ytest, predictions)
                        # n_alphas_batch, n_targets_batch = score.shape
                    del Ytrain, Ytest

                # make small alphas impossible to select
                too_small_alphas = backend.isnan(matrix[:, 0, 0])
                scores[jj, alpha_batch, :][too_small_alphas] = -1e5

                del matrix, predictions
            del train, test, Xtrain, Xtest

        # select best alphas
        alphas_argmax, cv_scores_ii = _select_best_alphas(
            scores, alphas, local_alpha, conservative)
        cv_scores[ii, :] = backend.to_cpu(cv_scores_ii)

        # update best_gammas and best_alphas
        epsilon = np.finfo(_dtype_to_str(dtype)).eps
        mask = cv_scores_ii > current_best_scores + epsilon
        current_best_scores[mask] = cv_scores_ii[mask]
        best_gammas[:, mask] = gamma[:, None]
        best_alphas[mask] = alphas[alphas_argmax[mask]]

        # compute primal or dual weights on the entire dataset (nocv)
        if return_weights:
            update_indices = backend.flatnonzero(mask)
            if Y_in_cpu:
                update_indices = backend.to_cpu(update_indices)
            if len(update_indices) > 0:

                # refit weights only for alphas used by at least one target
                used_alphas = backend.unique(best_alphas[mask])
                primal_weights = backend.zeros_like(
                    X_, shape=(n_features, len(update_indices)), device="cpu")
                for matrix, alpha_batch in _decompose_ridge(
                        Xtrain=X_, alphas=used_alphas,
                        negative_eigenvalues="zeros",
                        n_alphas_batch=min(len(used_alphas), n_alphas_batch),
                        method=diagonalize_method):

                    for start in range(0, len(update_indices),
                                       n_targets_batch_refit):
                        batch = slice(start, start + n_targets_batch_refit)

                        weights = backend.matmul(
                            matrix,
                            backend.to_gpu(Y[:, update_indices[batch]],
                                           device=device))
                        # used_n_alphas_batch, n_features, n_targets_batch = \
                        # weights.shape

                        # select alphas corresponding to best cv_score
                        alphas_indices = backend.searchsorted(
                            used_alphas, best_alphas[mask][batch])
                        # mask targets whose selected alphas are outside the
                        # alpha batch
                        mask2 = backend.isin(
                            alphas_indices,
                            backend.arange(len(used_alphas))[alpha_batch])
                        # get indices in alpha_batch
                        alphas_indices = backend.searchsorted(
                            backend.arange(len(used_alphas))[alpha_batch],
                            alphas_indices[mask2])
                        # update corresponding weights
                        tmp = weights[alphas_indices, :,
                                      backend.arange(weights.shape[2])[mask2]]
                        primal_weights[:, batch][:, backend.to_cpu(mask2)] = \
                            backend.to_cpu(tmp).T
                        del weights, alphas_indices, mask2
                    del matrix

                # multiply again by np.sqrt(g), as we then want to use
                # the primal weights on the unscaled features Xs, and not
                # on the scaled features (np.sqrt(g) * Xs)
                for kk in range(n_spaces):
                    primal_weights[slices[kk]] *= backend.to_cpu(
                        backend.sqrt(gamma[kk]))
                refit_weights[:, backend.to_cpu(mask)] = primal_weights
                del primal_weights

            del update_indices
        del mask

        for kk in range(n_spaces):
            X_[:, slices[kk]] /= backend.sqrt(gamma[kk])

    deltas = backend.log(best_gammas / best_alphas[None, :])

    if fit_intercept:
        intercept = (backend.to_cpu(Y_offset) -
                     backend.to_cpu(X_offset) @ refit_weights
                     ) if return_weights else None
        return deltas, refit_weights, cv_scores, intercept
    else:
        return deltas, refit_weights, cv_scores


def _decompose_ridge(Xtrain, alphas, n_alphas_batch=None, method="svd",
                     negative_eigenvalues="zeros"):
    """Precompute resolution matrices for ridge predictions.

    To compute the prediction::

        Ytest_hat = Xtest @ (XTX + alphas * Id)^-1 @ Xtrain^T @ Ytrain

        where XTX = Xtrain^T @ Xtrain,

    this function precomputes::

        matrices = (XTX + alphas * Id)^-1 @ Xtrain^T.

    Parameters
    ----------
    Xtrain : array of shape (n_samples_train, n_features)
        Concatenated input features.
    alphas : float, or array of shape (n_alphas, )
        Range of ridge regularization parameter.
    n_alphas_batch : int or None
        If not None, returns a generator over batches of alphas.
    method : str in {"svd"}
        Method used to diagonalize the kernel.
    negative_eigenvalues : str in {"nan", "error", "zeros"}
        If the decomposition leads to negative eigenvalues (wrongly emerging
        from float32 errors):
            - "error" raises an error.
            - "zeros" remplaces them with zeros.
            - "nan" returns nans if the regularization does not compensate
                twice the smallest negative value, else it ignores the problem.

    Returns
    -------
    matrices : array of shape (n_alphas, n_samples_test, n_samples_train) or \
        (n_alphas, n_features, n_samples_train) if test is not None
        Precomputed resolution matrices.
    alpha_batch : slice
        Slice of the batch of alphas.
    """
    backend = get_backend()

    use_alpha_batch = n_alphas_batch is not None
    if n_alphas_batch is None:
        n_alphas_batch = len(alphas)

    if method == "svd":
        # SVD: X = U @ np.diag(eigenvalues) @ Vt
        U, eigenvalues, Vt = backend.svd(Xtrain, full_matrices=False)
    else:
        raise ValueError("Unknown method=%r." % (method, ))

    for start in range(0, len(alphas), n_alphas_batch):
        batch = slice(start, start + n_alphas_batch)

        ev_weighting = eigenvalues / (alphas[batch, None] + eigenvalues ** 2)

        # negative eigenvalues can emerge from incorrect kernels,
        # or from float32
        if eigenvalues[0] < 0:
            if negative_eigenvalues == "nan":
                ev_weighting[alphas[batch] < -eigenvalues[0] * 2, :] = \
                    backend.asarray(backend.nan, type=ev_weighting.dtype)

            elif negative_eigenvalues == "zeros":
                eigenvalues[eigenvalues < 0] = 0

            elif negative_eigenvalues == "error":
                raise RuntimeError(
                    "Negative eigenvalues. Make sure the kernel is positive "
                    "semi-definite, increase the regularization alpha, or use"
                    "another solver.")
            else:
                raise ValueError("Unknown negative_eigenvalues=%r." %
                                 (negative_eigenvalues, ))

        matrices = backend.matmul(Vt.T, ev_weighting[:, :, None] * U.T)

        if use_alpha_batch:
            yield matrices, batch
        else:
            return matrices, batch

        del matrices


#: Dictionary with all group ridge solvers
GROUP_RIDGE_SOLVERS = {
    "random_search": solve_group_ridge_random_search,
}


def solve_ridge_cv_svd(X, Y, alphas=1.0, fit_intercept=False,
                       score_func=l2_neg_loss, cv=5, local_alpha=True,
                       n_targets_batch=None, n_targets_batch_refit=None,
                       n_alphas_batch=None, conservative=False,
                       Y_in_cpu=False):
    """Solve ridge regression with a grid search over alphas.

    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        Input features.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alphas : float or array of shape (n_alphas, )
        Range of ridge regularization parameter.
    fit_intercept : boolean
        Whether to fit an intercept.
        If False, X and Y must be zero-mean over samples.
    score_func : callable
        Function used to compute the score of predictions versus Y.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    local_alpha : bool
        If True, alphas are selected per target, else shared over all targets.
    n_targets_batch : int or None
        Size of the batch for over targets during cross-validation.
        Used for memory reasons. If None, uses all n_targets at once.
    n_targets_batch_refit : int or None
        Size of the batch for over targets during refit.
        Used for memory reasons. If None, uses all n_targets at once.
    n_alphas_batch : int or None
        Size of the batch for over alphas. Used for memory reasons.
        If None, uses all n_alphas at once.
    conservative : bool
        If True, when selecting the hyperparameter alpha, take the largest one
        that is less than one standard deviation away from the best.
        If False, take the best.
    Y_in_cpu : bool
        If True, keep the target values ``Y`` in CPU memory (slower).

    Returns
    -------
    best_alphas : array of shape (n_targets, )
        Selected best hyperparameter alphas.
    coefs : array of shape (n_samples, n_targets)
        Ridge coefficients refit on the entire dataset, using selected
        best hyperparameters alpha. Always stored on CPU memory.
    cv_scores : array of shape (n_targets, )
        Cross-validation scores averaged over splits, for the best alpha.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()

    n_iter = backend.ones_like(X, shape=(1, 1))
    fixed_params = dict(return_weights=True, progress_bar=False,
                        concentration=None, jitter_alphas=False,
                        random_state=None, n_iter=n_iter)

    copied_params = dict(alphas=alphas, score_func=score_func, cv=cv,
                         local_alpha=local_alpha, fit_intercept=fit_intercept,
                         n_targets_batch=n_targets_batch,
                         n_targets_batch_refit=n_targets_batch_refit,
                         n_alphas_batch=n_alphas_batch,
                         conservative=conservative, Y_in_cpu=Y_in_cpu)

    tmp = solve_group_ridge_random_search([X], Y, **copied_params,
                                          **fixed_params)

    if fit_intercept:
        deltas, coefs, cv_scores, intercept = tmp
        best_alphas = backend.exp(-deltas[0])
        return best_alphas, coefs, cv_scores, intercept
    else:
        deltas, coefs, cv_scores = tmp
        best_alphas = backend.exp(-deltas[0])
        return best_alphas, coefs, cv_scores



from sklearn.base import BaseEstimator, RegressorMixin, MultiOutputMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.model_selection import check_cv

from ._group_lasso import solve_sparse_group_lasso_cv

from ..validation import check_array
from ..validation import _get_string_dtype
from ..backend import get_backend
from ..scoring import r2_score


class SparseGroupLassoCV(MultiOutputMixin, RegressorMixin, BaseEstimator):
    """Sparse group Lasso

    Solved with hyperparameter grid-search over cross-validation.

    Parameters
    ----------
    groups : array of shape (n_features, ) or None
        Encoding of the group of each feature. If None, all features are
        gathered in one group, and the problem is equivalent to the Lasso.

    l21_regs : array of shape (n_l21_regs, )
        All the Group Lasso regularization parameter tested.

    l1_regs : array of shape (n_l1_regs, )
        All the Lasso regularization parameter tested.

    solver : str
        Algorithm used during the fit, "proximal_gradient" only for now.

    solver_params : dict or None
        Additional parameters for the solver.
        See more details in the docstring of the function:
        ``SparseGroupLassoCV.ALL_SOLVERS[solver]``

    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.

    Attributes
    ----------
    coef_ : array of shape (n_samples) or (n_samples, n_targets)
        Coefficient of the linear model. Always on CPU.

    best_l21_reg_ : array of shape (n_targets, )
        Best hyperparameter per target.

    best_l1_reg_ : array of shape (n_targets, )
        Best hyperparameter per target.

    cv_scores_ : array of shape (n_l21_regs * n_l1_regs, n_targets)
        Cross-validation scores of all tested hyperparameters.

    n_features_in_ : int
        Number of features used during the fit.

    Examples
    --------
    >>> from himalaya.lasso import SparseGroupLassoCV
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 3
    >>> X = np.random.randn(n_samples, n_features)
    >>> Y = np.random.randn(n_samples, n_targets)
    >>> clf = SparseGroupLassoCV()
    >>> clf.fit(X, Y)
    SparseGroupLassoCV()
    """
    ALL_SOLVERS = dict(proximal_gradient=solve_sparse_group_lasso_cv)

    def __init__(self, groups=None, l1_regs=[0], l21_regs=[0],
                 solver="proximal_gradient", solver_params=None, cv=5):
        self.groups = groups
        self.l1_regs = l1_regs
        self.l21_regs = l21_regs
        self.solver = solver
        self.solver_params = solver_params
        self.cv = cv

    def fit(self, X, y):
        """Fit the model

        Parameters
        ----------
        X : array of shape (n_samples, n_features).
            Training data.

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        X = check_array(X, accept_sparse=False, ndim=2)
        self.dtype_ = _get_string_dtype(X)
        y = check_array(y, dtype=self.dtype_, ndim=[1, 2])
        if X.shape[0] != y.shape[0]:
            raise ValueError("Inconsistent number of samples.")

        self.n_features_in_ = X.shape[1]
        cv = check_cv(self.cv)
        ravel = False
        if y.ndim == 1:
            y = y[:, None]
            ravel = True

        results = self._call_solver(X=X, Y=y, groups=self.groups, cv=cv,
                                    l21_regs=self.l21_regs,
                                    l1_regs=self.l1_regs)
        self.coef_, self.best_l21_reg_, self.best_l1_reg_ = results[:3]
        self.cv_scores_ = results[3]

        if ravel:
            self.coef_ = self.coef_[:, 0]

        return self

    def _call_solver(self, **direct_params):
        if self.solver not in self.ALL_SOLVERS:
            raise ValueError("Unknown solver=%r." % self.solver)

        function = self.ALL_SOLVERS[self.solver]
        solver_params = self.solver_params or {}

        # check duplicated parameters
        intersection = set(direct_params.keys()).intersection(
            set(solver_params.keys()))
        if intersection:
            raise ValueError(
                'Parameters %s should not be given in solver_params, since '
                'they are either fixed or have a direct parameter in %s.' %
                (intersection, self.__class__.__name__))

        return function(**direct_params, **solver_params)

    def predict(self, X):
        """Predict using the model.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features)
            Samples.

        Returns
        -------
        Y_hat : array of shape (n_samples,) or (n_samples, n_targets)
            Returns predicted values.
        """
        backend = get_backend()
        check_is_fitted(self)
        X = check_array(X, dtype=self.dtype_, accept_sparse=False, ndim=2)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                'Different number of features in X than during fit.')
        Y_hat = backend.to_numpy(X) @ backend.to_numpy(self.coef_)
        return backend.asarray_like(Y_hat, ref=X)

    def score(self, X, y):
        """Return the coefficient of determination R^2 of the prediction.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features)
            Samples.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            True values for X.

        Returns
        -------
        score : array of shape (n_targets, )
            R^2 of self.predict(X) versus y.
        """
        y_pred = self.predict(X, y)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return r2_score(y_true[:, None], y_pred[:, None])[0]
        else:
            return r2_score(y_true, y_pred)

    def _more_tags(self):
        return {'requires_y': True}



import warnings
import numbers

import numpy as np
from sklearn.model_selection import check_cv

from ..backend import get_backend
from ..backend._utils import _dtype_to_str
from ..progress_bar import bar
from ..scoring import l2_neg_loss
from ..validation import check_random_state
from ._solvers import _helper_intercept
from ._kernels import KernelCenterer


def solve_multiple_kernel_ridge_random_search(
        Ks, Y, n_iter=100, concentration=[0.1, 1.0], alphas=1.0,
        score_func=l2_neg_loss, cv=5, fit_intercept=False, return_weights=None,
        Xs=None, local_alpha=True, jitter_alphas=False, random_state=None,
        n_targets_batch=None, n_targets_batch_refit=None, n_alphas_batch=None,
        progress_bar=True, Ks_in_cpu=False, conservative=False, Y_in_cpu=False,
        diagonalize_method="eigh", return_alphas=False):
    """Solve multiple kernel ridge regression using random search.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Input kernels.
    Y : array of shape (n_samples, n_targets)
        Target data.
    n_iter : int, or array of shape (n_iter, n_kernels)
        Number of kernel weights combination to search.
        If an array is given, the solver uses it as the list of kernel weights
        to try, instead of sampling from a Dirichlet distribution.
    concentration : float, or list of float
        Concentration parameters of the Dirichlet distribution.
        If a list, iteratively cycle through the list.
        Not used if n_iter is an array.
    alphas : float or array of shape (n_alphas, )
        Range of ridge regularization parameter.
    score_func : callable
        Function used to compute the score of predictions versus Y.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    fit_intercept : boolean
        Whether to fit an intercept. If False, Ks should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
        Only available if return_weights == 'dual'.
    return_weights : None, 'primal', or 'dual'
        Whether to refit on the entire dataset and return the weights.
    Xs : array of shape (n_kernels, n_samples, n_features) or None
        Necessary if return_weights == 'primal'.
    local_alpha : bool
        If True, alphas are selected per target, else shared over all targets.
    jitter_alphas : bool
        If True, alphas range is slightly jittered for each gamma.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.
    n_targets_batch : int or None
        Size of the batch for over targets during cross-validation.
        Used for memory reasons. If None, uses all n_targets at once.
    n_targets_batch_refit : int or None
        Size of the batch for over targets during refit.
        Used for memory reasons. If None, uses all n_targets at once.
    n_alphas_batch : int or None
        Size of the batch for over alphas. Used for memory reasons.
        If None, uses all n_alphas at once.
    progress_bar : bool
        If True, display a progress bar over gammas.
    Ks_in_cpu : bool
        If True, keep Ks in CPU memory to limit GPU memory (slower).
        This feature is not available through the scikit-learn API.
    conservative : bool
        If True, when selecting the hyperparameter alpha, take the largest one
        that is less than one standard deviation away from the best.
        If False, take the best.
    Y_in_cpu : bool
        If True, keep the target values ``Y`` in CPU memory (slower).
    diagonalize_method : str in {"eigh", "svd"}
        Method used to diagonalize the kernel.
    return_alphas : bool
        If True, return the best alpha value for each target.

    Returns
    -------
    deltas : array of shape (n_kernels, n_targets)
        Best log kernel weights for each target.
    refit_weights : array or None
        Refit regression weights on the entire dataset, using selected best
        hyperparameters. Refit weights are always stored on CPU memory.
        If return_weights == 'primal', shape is (n_features, n_targets),
        if return_weights == 'dual', shape is (n_samples, n_targets),
        else, None.
    cv_scores : array of shape (n_iter, n_targets)
        Cross-validation scores per iteration, averaged over splits, for the
        best alpha. Cross-validation scores will always be on CPU memory.
    best_alphas : array of shape (n_targets, )
        Best alpha value per target. Only returned if return_alphas is True.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    if isinstance(n_iter, int):
        gammas = generate_dirichlet_samples(n_samples=n_iter,
                                            n_kernels=len(Ks),
                                            concentration=concentration,
                                            random_state=random_state)
        gammas[0] = 1 / len(Ks)
    elif n_iter.ndim == 2:
        gammas = n_iter
        assert gammas.shape[1] == Ks.shape[0]
    else:
        raise ValueError("Unknown parameter n_iter=%r." % (n_iter, ))

    if isinstance(alphas, numbers.Number) or alphas.ndim == 0:
        alphas = backend.ones_like(Y, shape=(1, )) * alphas

    dtype = Ks.dtype
    gammas = backend.asarray(gammas, dtype=dtype)
    device = getattr(gammas, "device", None)
    gammas, alphas, Xs = backend.check_arrays(gammas, alphas, Xs)
    Y = backend.asarray(Y, dtype=dtype, device="cpu" if Y_in_cpu else device)
    Ks = backend.asarray(Ks, dtype=dtype,
                         device="cpu" if Ks_in_cpu else device)

    if fit_intercept:
        if return_weights == 'dual':
            Ks, Y, Ks_rows, Y_offset = _helper_intercept(Ks, Y)
        elif return_weights == 'primal':
            raise NotImplementedError(
                'Computing the intercept with return_weights="primal" is not'
                'implemented. Use fit_intercept=False or'
                'return_weights="dual".')
        else:
            raise ValueError(f'Cannot compute the intercept if return_weights'
                             f'is equal to {return_weights}.')

    n_samples, n_targets = Y.shape
    if n_targets_batch is None:
        n_targets_batch = n_targets
    if n_targets_batch_refit is None:
        n_targets_batch_refit = n_targets_batch
    if n_alphas_batch is None:
        n_alphas_batch = len(alphas)

    cv = check_cv(cv)
    n_splits = cv.get_n_splits()
    n_kernels = len(Ks)
    for train, val in cv.split(Y):
        if len(val) == 0 or len(train) == 0:
            raise ValueError("Empty train or validation set. "
                             "Check that `cv` is correctly defined.")

    if jitter_alphas:
        random_generator = check_random_state(random_state)
        given_alphas = backend.copy(alphas)

    best_gammas = backend.full_like(gammas, fill_value=1.0 / n_kernels,
                                    shape=(n_kernels, n_targets))
    best_alphas = backend.ones_like(gammas, shape=n_targets)
    cv_scores = backend.zeros_like(gammas, shape=(len(gammas), n_targets),
                                   device="cpu")
    current_best_scores = backend.full_like(gammas, fill_value=-backend.inf,
                                            shape=n_targets)

    # initialize refit ridge weights
    if return_weights == 'primal':
        if Xs is None:
            raise ValueError("Xs is needed to compute the primal weights.")
        n_features = sum(X.shape[1] for X in Xs)
        refit_weights = backend.zeros_like(gammas,
                                           shape=(n_features, n_targets),
                                           device="cpu")

    elif return_weights == 'dual':
        refit_weights = backend.zeros_like(gammas,
                                           shape=(n_samples, n_targets),
                                           device="cpu")
    elif return_weights is None:
        refit_weights = None
    else:
        raise ValueError("Unknown parameter return_weights=%r." %
                         (return_weights, ))

    for ii, gamma in enumerate(
            bar(gammas, '%d random sampling with cv' % len(gammas),
                use_it=progress_bar)):

        if Ks_in_cpu:
            K = (backend.to_cpu(gamma[:, None, None]) * Ks).sum(0)
            K = backend.to_gpu(K, device=device)
        else:
            K = (gamma[:, None, None] * Ks).sum(0)

        if jitter_alphas:
            noise = backend.asarray_like(random_generator.rand(), alphas)
            alphas = given_alphas * (10 ** (noise - 0.5))

        scores = backend.zeros_like(gammas,
                                    shape=(n_splits, len(alphas), n_targets))
        for jj, (train, test) in enumerate(cv.split(K)):
            train = backend.to_gpu(train, device=device)
            test = backend.to_gpu(test, device=device)
            Ktrain, Ktest = K[train[:, None], train], K[test[:, None], train]
            if fit_intercept:
                centerer = KernelCenterer()
                Ktrain = centerer.fit_transform(Ktrain)
                Ktest = centerer.transform(Ktest)

            for matrix, alpha_batch in _decompose_kernel_ridge(
                    Ktrain=Ktrain, alphas=alphas, Ktest=Ktest,
                    negative_eigenvalues="nan", n_alphas_batch=n_alphas_batch,
                    method=diagonalize_method):
                # n_alphas_batch, n_samples_test, n_samples_train = \
                # matrix.shape

                for start in range(0, n_targets, n_targets_batch):
                    batch = slice(start, start + n_targets_batch)
                    Ybatch = backend.to_gpu(Y[:, batch], device=device)
                    Ytrain, Ytest = Ybatch[train], Ybatch[test]
                    del Ybatch
                    if fit_intercept:
                        Ytrain_mean = Ytrain.mean(0)
                        Ytrain = Ytrain - Ytrain_mean
                        Ytest = Ytest - Ytrain_mean

                    predictions = backend.matmul(matrix, Ytrain)
                    # n_alphas_batch, n_samples_test, n_targets_batch = \
                    # predictions.shape

                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore", category=UserWarning)
                        scores[jj, alpha_batch, batch] = score_func(
                            Ytest, predictions)
                        # n_alphas_batch, n_targets_batch = score.shape

                # make small alphas impossible to select
                too_small_alphas = backend.isnan(matrix[:, 0, 0])
                scores[jj, alpha_batch, :][too_small_alphas] = -1e5

                del matrix, predictions
            del train, test

        # select best alphas
        alphas_argmax, cv_scores_ii = _select_best_alphas(
            scores, alphas, local_alpha, conservative)
        cv_scores[ii, :] = backend.to_cpu(cv_scores_ii)

        # update best_gammas and best_alphas
        epsilon = np.finfo(_dtype_to_str(dtype)).eps
        mask = cv_scores_ii > current_best_scores + epsilon
        current_best_scores[mask] = cv_scores_ii[mask]
        best_gammas[:, mask] = gamma[:, None]
        best_alphas[mask] = alphas[alphas_argmax[mask]]

        # compute primal or dual weights on the entire dataset (nocv)
        if return_weights is not None:
            update_indices = backend.flatnonzero(mask)
            if Y_in_cpu:
                update_indices = backend.to_cpu(update_indices)
            if len(update_indices) > 0:

                # refit weights only for alphas used by at least one target
                used_alphas = backend.unique(best_alphas[mask])
                dual_weights = backend.zeros_like(
                    K, shape=(n_samples, len(update_indices)), device="cpu")
                for matrix, alpha_batch in _decompose_kernel_ridge(
                        K, used_alphas, Ktest=None,
                        negative_eigenvalues="zeros",
                        n_alphas_batch=min(len(used_alphas), n_alphas_batch),
                        method=diagonalize_method):

                    for start in range(0, len(update_indices),
                                       n_targets_batch_refit):
                        batch = slice(start, start + n_targets_batch_refit)

                        weights = backend.matmul(
                            matrix,
                            backend.to_gpu(Y[:, update_indices[batch]],
                                           device=device))
                        # used_n_alphas_batch, n_samples, n_targets_batch = \
                        # weights.shape

                        # select alphas corresponding to best cv_score
                        alphas_indices = backend.searchsorted(
                            used_alphas, best_alphas[mask][batch])
                        # mask targets whose selected alphas are outside the
                        # alpha batch
                        mask2 = backend.isin(
                            alphas_indices,
                            backend.arange(len(used_alphas))[alpha_batch])
                        # get indices in alpha_batch
                        alphas_indices = backend.searchsorted(
                            backend.arange(len(used_alphas))[alpha_batch],
                            alphas_indices[mask2])
                        # update corresponding weights
                        tmp = weights[alphas_indices, :,
                                      backend.arange(weights.shape[2])[mask2]]
                        dual_weights[:, batch][:, backend.to_cpu(mask2)] = \
                            backend.to_cpu(tmp).T
                        del weights, alphas_indices, mask2
                    del matrix

                if return_weights == 'primal':
                    # multiply by g and not np.sqrt(g), as we then want to use
                    # the primal weights on the unscaled features Xs, and not
                    # on the scaled features (np.sqrt(g) * Xs)
                    X = backend.concatenate([t * g for t, g in zip(Xs, gamma)],
                                            1)
                    primal_weights = backend.to_cpu(X.T) @ dual_weights
                    refit_weights[:, backend.to_cpu(mask)] = primal_weights
                    del X, primal_weights

                elif return_weights == 'dual':
                    refit_weights[:, backend.to_cpu(mask)] = dual_weights

                del dual_weights
            del update_indices
        del K, mask

    deltas = backend.log(best_gammas / best_alphas[None, :])
    if return_weights == 'dual':
        refit_weights *= backend.to_cpu(best_alphas)

    if fit_intercept:
        if return_weights == 'dual':
            intercept = backend.to_cpu(Y_offset)
            intercept -= ((backend.to_cpu(Ks_rows) @ refit_weights) *
                          backend.to_cpu(backend.exp(deltas))).sum(0)
        else:
            return NotImplementedError()

    results = [deltas, refit_weights, cv_scores]
    if return_alphas:
        results.append(best_alphas)
    if fit_intercept:
        results.append(intercept)
    return results


def _select_best_alphas(scores, alphas, local_alpha, conservative):
    """Helper to select the best alphas

    Parameters
    ----------
    scores : array of shape (n_splits, n_alphas, n_targets)
        Cross validation scores.
    alphas :array of shape (n_alphas, )
        Regularization parameters.
    local_alpha : bool
        If True, alphas are selected per target, else shared over all targets.
    conservative : bool
        If True, when selecting the hyperparameter alpha, take the largest one
        that is less than one standard deviation away from the best.
        If False, take the best.

    Returns
    -------
    alphas_argmax : array of shape (n_targets, )
        Indices of the best alphas.
    best_scores_mean : arrya of shape (n_targets, )
        Scores, averaged over splits, and maximized over alphas.
    """
    backend = get_backend()

    # average scores over splits
    scores_mean = backend.mean_float64(scores, axis=0)
    # add epsilon slope to select larger alphas if scores are equal
    scores_mean += (backend.log(alphas) * 1e-10)[:, None]

    # compute the max over alphas
    axis = 0
    if local_alpha:
        alphas_argmax = backend.argmax(scores_mean, axis)

        if conservative:
            # take a conservative alpha, the largest that beats the best
            # score minus one standard deviation
            scores_std = scores.std(0)
            to_beat = backend.apply_argmax(scores_mean - scores_std,
                                           alphas_argmax, axis)
            does_beat = backend.asarray(scores_mean > to_beat, dtype="float32")
            # add bias toward large alphas (scaled to be << 1)
            does_beat += backend.log(alphas)[:, None] * 1e-4
            alphas_argmax = backend.argmax(does_beat, axis)
    else:
        if conservative:
            raise NotImplementedError()
        else:
            alphas_argmax = backend.argmax(scores_mean.mean(1))
            alphas_argmax = backend.full_like(alphas_argmax,
                                              shape=scores_mean.shape[1],
                                              fill_value=alphas_argmax)
    best_scores_mean = backend.apply_argmax(scores_mean, alphas_argmax, axis)

    return alphas_argmax, best_scores_mean


def generate_dirichlet_samples(n_samples, n_kernels, concentration=[.1, 1.],
                               random_state=None):
    """Generate samples from a Dirichlet distribution.

    Parameters
    ----------
    n_samples : int
        Number of samples to generate.
    n_kernels : int
        Number of dimension of the distribution.
    concentration : float, or list of float
        Concentration parameters of the Dirichlet distribution.
        A value of 1 corresponds to uniform sampling over the simplex.
        A value of infinity corresponds to equal weights.
        If a list, samples cycle through the list.
    random_state : int, or None
        Random generator seed. Use an int for deterministic samples.

    Returns
    -------
    gammas : array of shape (n_samples, n_kernels)
        Dirichlet samples.
    """
    import numpy as np
    random_generator = check_random_state(random_state)

    concentration = np.atleast_1d(concentration)
    n_concentrations = len(concentration)
    n_samples_per_concentration = int(
        np.ceil(n_samples / float(n_concentrations)))

    # generate the gammas
    gammas = []
    for conc in concentration:
        if conc == np.inf:
            gamma = np.full(n_kernels, fill_value=1. / n_kernels)[None]
            gamma = np.tile(gamma, (n_samples_per_concentration, 1))
        else:
            gamma = random_generator.dirichlet([conc] * n_kernels,
                                               n_samples_per_concentration)
        gammas.append(gamma)
    gammas = np.vstack(gammas)

    # reorder the gammas to alternate between concentrations:
    # [a0, a1, a2, a0, a1, a2] instead of [a0, a0, a1, a1, a2, a2]
    gammas = gammas.reshape(n_concentrations, n_samples_per_concentration,
                            n_kernels)
    gammas = np.swapaxes(gammas, 0, 1)
    gammas = gammas.reshape(n_concentrations * n_samples_per_concentration,
                            n_kernels)

    # remove extra gammas
    gammas = gammas[:n_samples]

    # cast to current backend
    backend = get_backend()
    gammas = backend.asarray(gammas)

    return gammas


def _decompose_kernel_ridge(Ktrain, alphas, Ktest=None, n_alphas_batch=None,
                            method="eigh", negative_eigenvalues="zeros"):
    """Precompute resolution matrices for kernel ridge predictions.

    To compute the prediction::

        Ytest_hat = Ktest @ (Ktrain + alphas * Id)^-1 @ Ytrain

    this function precomputes::

        matrices = Ktest @ (Ktrain + alphas * Id)^-1

    or just::

        matrices = (Ktrain + alphas * Id)^-1

    if Ktest is None.

    Parameters
    ----------
    Ktrain : array of shape (n_samples_train, n_samples_train)
        Training kernel for one feature space.
    alphas : float, or array of shape (n_alphas, )
        Range of ridge regularization parameter.
    Ktest : array of shape (n_samples_test, n_samples_train)
        Testing kernel for one feature space.
    n_alphas_batch : int or None
        If not None, returns a generator over batches of alphas.
    method : str in {"eigh", "svd"}
        Method used to diagonalize the kernel.
    negative_eigenvalues : str in {"nan", "error", "zeros"}
        If the decomposition leads to negative eigenvalues (wrongly emerging
        from float32 errors):
            - "error" raises an error.
            - "zeros" remplaces them with zeros.
            - "nan" returns nans if the regularization does not compensate
                twice the smallest negative value, else it ignores the problem.

    Returns
    -------
    matrices : array of shape (n_alphas, n_samples_train, n_samples_train) or \
        (n_alphas, n_samples_test, n_samples_train) if Ktest is not None
        Precomputed resolution matrices.
    alpha_batch : slice
        Slice of the batch of alphas.
    """
    backend = get_backend()

    use_alpha_batch = n_alphas_batch is not None
    if n_alphas_batch is None:
        n_alphas_batch = len(alphas)

    if method == "eigh":
        # diagonalization: K = V @ np.diag(eigenvalues) @ V.T
        eigenvalues, V = backend.eigh(Ktrain)
        # match SVD notations: K = U @ np.diag(eigenvalues) @ Vt
        U = V
        Vt = V.T
    elif method == "svd":
        # SVD: K = U @ np.diag(eigenvalues) @ Vt
        U, eigenvalues, Vt = backend.svd(Ktrain)
    else:
        raise ValueError("Unknown method=%r." % (method, ))

    if Ktest is not None:
        Ktest_V = backend.matmul(Ktest, Vt.T)

    for start in range(0, len(alphas), n_alphas_batch):
        batch = slice(start, start + n_alphas_batch)

        ev_weighting = (alphas[batch, None] + eigenvalues) ** -1

        # negative eigenvalues can emerge from incorrect kernels,
        # or from float32
        if eigenvalues[0] < 0:
            if negative_eigenvalues == "nan":
                ev_weighting[alphas[batch] < -eigenvalues[0] *
                             2, :] = backend.asarray(backend.nan,
                                                     dtype=ev_weighting.dtype)

            elif negative_eigenvalues == "zeros":
                eigenvalues[eigenvalues < 0] = 0

            elif negative_eigenvalues == "error":
                raise RuntimeError(
                    "Negative eigenvalues. Make sure the kernel is positive "
                    "semi-definite, increase the regularization alpha, or use"
                    "another solver.")
            else:
                raise ValueError("Unknown negative_eigenvalues=%r." %
                                 (negative_eigenvalues, ))

        if Ktest is not None:
            matrices = backend.matmul(Ktest_V, ev_weighting[:, :, None] * U.T)
        else:
            matrices = backend.matmul(Vt.T, ev_weighting[:, :, None] * U.T)

        if use_alpha_batch:
            yield matrices, batch
        else:
            return matrices, batch

        del matrices


def solve_kernel_ridge_cv_eigenvalues(K, Y, alphas=1.0, score_func=l2_neg_loss,
                                      cv=5, fit_intercept=False,
                                      local_alpha=True, n_targets_batch=None,
                                      n_targets_batch_refit=None,
                                      n_alphas_batch=None, conservative=False,
                                      Y_in_cpu=False):
    """Solve kernel ridge regression with a grid search over alphas.

    Parameters
    ----------
    K : array of shape (n_samples, n_samples)
        Input kernel.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alphas : float or array of shape (n_alphas, )
        Range of ridge regularization parameter.
    score_func : callable
        Function used to compute the score of predictions versus Y.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    fit_intercept : boolean
        Whether to fit an intercept. If False, K should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
    local_alpha : bool
        If True, alphas are selected per target, else shared over all targets.
    n_targets_batch : int or None
        Size of the batch for over targets during cross-validation.
        Used for memory reasons. If None, uses all n_targets at once.
    n_targets_batch_refit : int or None
        Size of the batch for over targets during refit.
        Used for memory reasons. If None, uses all n_targets at once.
    n_alphas_batch : int or None
        Size of the batch for over alphas. Used for memory reasons.
        If None, uses all n_alphas at once.
    conservative : bool
        If True, when selecting the hyperparameter alpha, take the largest one
        that is less than one standard deviation away from the best.
        If False, take the best.
    Y_in_cpu : bool
        If True, keep the target values ``Y`` in CPU memory (slower).

    Returns
    -------
    best_alphas : array of shape (n_targets, )
        Selected best hyperparameter alphas.
    dual_weights : array of shape (n_samples, n_targets)
        Kernel ridge coefficients refit on the entire dataset, using selected
        best hyperparameters alpha. Always stored on CPU memory.
    cv_scores : array of shape (n_targets, )
        Cross-validation scores averaged over splits, for the best alpha.
    """
    backend = get_backend()

    n_iter = backend.ones_like(K, shape=(1, 1))
    fixed_params = dict(return_weights="dual", Xs=None, progress_bar=False,
                        concentration=None, jitter_alphas=False,
                        random_state=None, n_iter=n_iter)

    copied_params = dict(alphas=alphas, score_func=score_func, cv=cv,
                         local_alpha=local_alpha, fit_intercept=fit_intercept,
                         n_targets_batch=n_targets_batch,
                         n_targets_batch_refit=n_targets_batch_refit,
                         n_alphas_batch=n_alphas_batch,
                         conservative=conservative, Y_in_cpu=Y_in_cpu)

    tmp = solve_multiple_kernel_ridge_random_search(K[None], Y,
                                                    **copied_params,
                                                    **fixed_params)
    if fit_intercept:
        deltas, dual_weights, cv_scores, intercept = tmp
        best_alphas = backend.exp(-deltas[0])
        dual_weights /= backend.to_cpu(best_alphas)
        return best_alphas, dual_weights, cv_scores, intercept
    else:
        deltas, dual_weights, cv_scores = tmp
        best_alphas = backend.exp(-deltas[0])
        dual_weights /= backend.to_cpu(best_alphas)
        return best_alphas, dual_weights, cv_scores



import pytest

from himalaya.backend import set_backend
from himalaya.backend import get_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.backend._utils import _dtype_to_str
from himalaya.utils import assert_array_almost_equal


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_set_backend_correct(backend):
    # test the change of backend
    module = set_backend(backend)
    assert module.__name__.split('.')[-1] == backend

    # test idempotence
    module = set_backend(set_backend(backend))
    assert module.__name__.split('.')[-1] == backend

    # test set and get
    module = set_backend(get_backend())
    assert module.__name__.split('.')[-1] == backend

    assert set_backend(backend)


def test_set_backend_incorrect():
    for backend in ["wrong", ["numpy"], True, None, 10]:
        with pytest.raises(ValueError):
            set_backend(backend)
        with pytest.raises(ValueError):
            set_backend(backend, on_error="raise")
        with pytest.warns(Warning):
            set_backend(backend, on_error="warn")
        with pytest.raises(ValueError):
            set_backend(backend, on_error="foo")


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_apply_argmax(backend):
    backend = set_backend(backend)
    for array in [
            backend.randn(1),
            backend.randn(10),
            backend.randn(10, 1),
            backend.randn(10, 4),
            backend.randn(10, 1, 8),
            backend.randn(10, 4, 8),
    ]:
        for axis in range(array.ndim):
            argmax = backend.argmax(array, axis=axis)
            assert_array_almost_equal(
                backend.max(array, axis=axis),
                backend.apply_argmax(array, argmax, axis=axis),
            )


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_std_float64(backend, dtype_str):
    backend = set_backend(backend)
    for array in [
            backend.randn(1),
            backend.randn(10),
            backend.randn(10, 1),
            backend.randn(10, 4),
            backend.randn(10, 1, 8),
            backend.randn(10, 4, 8),
    ]:
        array = backend.asarray(array, dtype=dtype_str)
        array_64 = backend.asarray(array, dtype="float64")
        for axis in range(array.ndim):
            result = backend.std_float64(array, axis=axis)
            reference = backend.to_numpy(array_64).std(axis=axis,
                                                       dtype="float64")
            reference = backend.asarray(reference, dtype=dtype_str)
            assert_array_almost_equal(result, reference)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_diagonal_view(backend):
    backend = set_backend(backend)
    try:
        import torch
    except ImportError:
        pytest.skip("PyTorch not installed.")
    import numpy as np

    for array in [
            backend.randn(10, 4),
            backend.randn(10, 4).T,
            backend.randn(10, 4, 8),
            backend.randn(10, 4, 8).T,
            backend.randn(3, 4, 8, 5),
    ]:
        for axis1 in range(array.ndim):
            for axis2 in range(array.ndim):
                if axis1 != axis2:
                    result = backend.diagonal_view(array, axis1=axis1,
                                                   axis2=axis2)
                    # compare with torch diagonal
                    reference = torch.diagonal(
                        torch.from_numpy(backend.to_numpy(array)), dim1=axis1,
                        dim2=axis2)
                    assert_array_almost_equal(result, reference)
                    # compare with numpy diagonal
                    reference = np.diagonal(backend.to_numpy(array),
                                            axis1=axis1, axis2=axis2)
                    assert_array_almost_equal(result, reference)
                    # test that this is a modifiable view
                    result += 1
                    reference = np.diagonal(backend.to_numpy(array),
                                            axis1=axis1, axis2=axis2)
                    assert_array_almost_equal(result, reference)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_eigh(backend):
    import scipy.linalg
    backend = set_backend(backend)

    array = backend.randn(10, 20)
    array = backend.asarray(array, dtype='float64')
    kernel = array @ array.T

    values, vectors = backend.eigh(kernel)
    values_ref, vectors_ref = scipy.linalg.eigh(backend.to_numpy(kernel))

    assert_array_almost_equal(values, values_ref)

    # vectors can be flipped in sign
    assert vectors.shape == vectors_ref.shape
    for ii in range(vectors.shape[1]):
        try:
            assert_array_almost_equal(vectors[:, ii], vectors_ref[:, ii])
        except AssertionError:
            assert_array_almost_equal(vectors[:, ii], -vectors_ref[:, ii])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
@pytest.mark.parametrize('full_matrices', [True, False])
@pytest.mark.parametrize('three_dim', [True, False])
def test_svd(backend, full_matrices, three_dim):
    import numpy.linalg
    backend = set_backend(backend)

    if three_dim:
        array = backend.randn(3, 5, 7)
    else:
        array = backend.randn(5, 7)

    array = backend.asarray(array, dtype='float64')

    U, s, V = backend.svd(array, full_matrices=full_matrices)
    U_ref, s_ref, V_ref = numpy.linalg.svd(backend.to_numpy(array),
                                           full_matrices=full_matrices)

    assert_array_almost_equal(s, s_ref)

    if not three_dim:
        U_ref = U_ref[None]
        U = U[None]
        V_ref = V_ref[None]
        V = V[None]

    # vectors can be flipped in sign
    assert U.shape == U_ref.shape
    assert V.shape == V_ref.shape
    for kk in range(U.shape[0]):
        for ii in range(U.shape[2]):
            try:
                assert_array_almost_equal(U[kk, :, ii], U_ref[kk, :, ii])
                assert_array_almost_equal(V[kk, ii, :], V_ref[kk, ii, :])
            except AssertionError:
                assert_array_almost_equal(U[kk, :, ii], -U_ref[kk, :, ii])
                assert_array_almost_equal(V[kk, ii, :], -V_ref[kk, ii, :])


@pytest.mark.parametrize('backend_out', ALL_BACKENDS)
@pytest.mark.parametrize('backend_in', ALL_BACKENDS)
def test_changed_backend_asarray(backend_in, backend_out):
    backend = set_backend(backend_in)
    array_in = backend.asarray([1.2, 2.4, 4.8])
    assert array_in is not None

    # change the backend, and cast to the correct class
    backend = set_backend(backend_out)
    array_out = backend.asarray(array_in)
    assert array_out is not None

    if backend_in == backend_out or backend_in[:5] == backend_out[:5]:
        # assert the class did not change
        assert array_in.__class__ == array_out.__class__
    else:
        # assert the class did change
        assert array_in.__class__ != array_out.__class__

    # assert the new class is correct
    array_out2 = backend.randn(3)
    assert array_out.__class__ == array_out2.__class__

    # test check_arrays
    array_out3, array_out4, array_out5 = backend.check_arrays(
        array_in, array_in, [array_in])
    assert array_out.__class__ == array_out3.__class__
    assert array_out.__class__ == array_out4.__class__
    assert array_out.__class__ == array_out5[0].__class__


@pytest.mark.parametrize('dtype_out', ["float32", "float64"])
@pytest.mark.parametrize('dtype_in', ["float32", "float64"])
@pytest.mark.parametrize('backend_out', ALL_BACKENDS)
@pytest.mark.parametrize('backend_in', ALL_BACKENDS)
def test_asarray_dtype(backend_in, backend_out, dtype_in, dtype_out):
    backend = set_backend(backend_in)
    array_in = backend.asarray([1.2, 2.4, 4.8], dtype=dtype_in)
    assert _dtype_to_str(array_in.dtype) == dtype_in

    backend = set_backend(backend_out)
    array_out = backend.asarray(array_in, dtype=dtype_out)
    assert _dtype_to_str(array_out.dtype) == dtype_out


def test_dtype_to_str_wrong_input():
    assert _dtype_to_str(None) is None

    with pytest.raises(NotImplementedError):
        _dtype_to_str(42)



from sklearn.compose import ColumnTransformer
from sklearn.pipeline import _name_estimators


class ColumnTransformerNoStack(ColumnTransformer):
    """Applies transformers to columns of an array, and does not stack them.

    This estimator allows different columns or column subsets of the input to
    be transformed separately. The different groups of features generated are
    *not* stacked together, to be used e.g. in a GroupRidgeCV(groups="auto").
    This is useful to perform separate transformations on different feature
    spaces.

    Warning : This class does not perfectly abide by scikit-learn's API.
    Indeed, it returns a list of ``n_groups`` matrices of shape (n_samples,
    n_features_i), while scikit-learn's API only allows arrays of shape
    (n_samples, n_features). This class is intended to be used in a
    scikit-learn pipeline *just* before a GroupRidgeCV(groups="auto").

    Parameters
    ----------
    transformers : list of tuples
        List of (name, transformer, columns) tuples specifying the
        transformer objects to be applied to subsets of the data.

        name : str
            Like in Pipeline and FeatureUnion, this allows the transformer and
            its parameters to be set using ``set_params`` and searched in grid
            search.
        transformer : {'drop', 'passthrough'} or estimator
            Estimator must support ``fit`` and ``transform``.
            Special-cased strings 'drop' and 'passthrough' are accepted as
            well, to indicate to drop the columns or to pass them through
            untransformed, respectively.
        columns :  str, array-like of str, int, array-like of int, \
                array-like of bool, slice or callable
            Indexes the data on its second axis. Integers are interpreted as
            positional columns, while strings can reference DataFrame columns
            by name.  A scalar string or int should be used where
            ``transformer`` expects X to be a 1d array-like (vector),
            otherwise a 2d array will be passed to the transformer.
            A callable is passed the input data `X` and can return any of the
            above. To select multiple columns by name or dtype, you can use
            :obj:`make_column_selector`.

    remainder : {'drop', 'passthrough'} or estimator, default='drop'
        By default, only the specified columns in `transformers` are
        transformed and combined in the output, and the non-specified
        columns are dropped. (default of ``'drop'``).
        By specifying ``remainder='passthrough'``, all remaining columns that
        were not specified in `transformers` will be automatically passed
        through. This subset of columns is concatenated with the output of
        the transformers.
        By setting ``remainder`` to be an estimator, the remaining
        non-specified columns will use the ``remainder`` estimator. The
        estimator must support ``fit`` and ``transform``.
        Note that using this feature requires that the DataFrame columns
        input at ``fit`` and ``transform`` have identical order.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        n_jobs does not work with GPU backends.

    transformer_weights : dict, default=None
        Multiplicative weights for features per transformer. The output of the
        transformer is multiplied by these weights. Keys are transformer names,
        values the weights.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.

    Attributes
    ----------
    transformers_ : list
        The collection of fitted transformers as tuples of
        (name, fitted_transformer, column). `fitted_transformer` can be an
        estimator, 'drop', or 'passthrough'. In case there were no columns
        selected, this will be the unfitted transformer.
        If there are remaining columns, the final element is a tuple of the
        form:
        ('remainder', transformer, remaining_columns) corresponding to the
        ``remainder`` parameter. If there are remaining columns, then
        ``len(transformers_)==len(transformers)+1``, otherwise
        ``len(transformers_)==len(transformers)``.

    named_transformers_ : :class:`~sklearn.utils.Bunch`
        Read-only attribute to access any transformer by given name.
        Keys are transformer names and values are the fitted transformer
        objects.

    n_features_in_ : int
        Number of features used during the fit.

    sparse_output_ : False

    Notes
    -----
    The order of the columns in the transformed feature matrix follows the
    order of how the columns are specified in the `transformers` list.
    Columns of the original feature matrix that are not specified are
    dropped from the resulting transformed feature matrix, unless specified
    in the `passthrough` keyword. Those columns specified with `passthrough`
    are added at the right to the output of the transformers.

    See also
    --------
    himalaya.ridge.make_column_transform_no_stack : convenience function for
        combining the outputs of multiple pipelines applied to
        column subsets of the original feature space.
    sklearn.compose.make_column_selector : convenience function for selecting
        columns based on datatype or the columns name with a regex pattern.

    Examples
    --------
    >>> import numpy as np
    >>> from himalaya.ridge import ColumnTransformerNoStack
    >>> from sklearn.preprocessing import StandardScaler
    >>> ct = ColumnTransformerNoStack(
    ...     [("group_1", StandardScaler(), [0, 1, 2]),
    ...      ("group_2", StandardScaler(), slice(3, 5))])
    >>> X = np.random.randn(10, 5)
    >>> # Group separately the first three columns and the last two
    >>> # columns, creating two feature spaces.
    >>> Xs = ct.fit_transform(X)
    >>> print(Xi.shape for Xi in Xs)
    (2, 4, 4)
    """

    def _hstack(self, Xs):
        """Do *not* stack the feature spaces.

        In ColumnTransformer, this methods stacks Xs horizontally.
        Here instead, we return the list of Xs.

        Parameters
        ----------
        Xs : list of arrays of shape (n_samples, n_features_i)
        """
        return Xs


def make_column_transformer_no_stack(*transformers, **kwargs):
    """Construct a ColumnTransformerNoStack from the given transformers.

    This is a shorthand for the ColumnTransformerNoStack constructor; it does
    not require, and does not permit, naming the transformers. Instead, they
    will be given names automatically based on their types. It also does not
    allow weighting with ``transformer_weights``.

    Parameters
    ----------
    *transformers : tuples
        Tuples of the form (transformer, columns) specifying the
        transformer objects to be applied to subsets of the data.

        transformer : {'drop', 'passthrough'} or estimator
            Estimator must support ``fit`` and ``transform``.
            Special-cased strings 'drop' and 'passthrough' are accepted as
            well, to indicate to drop the columns or to pass them through
            untransformed, respectively.
        columns : str,  array-like of str, int, array-like of int, slice, \
                array-like of bool or callable
            Indexes the data on its second axis. Integers are interpreted as
            positional columns, while strings can reference DataFrame columns
            by name. A scalar string or int should be used where
            ``transformer`` expects X to be a 1d array-like (vector),
            otherwise a 2d array will be passed to the transformer.
            A callable is passed the input data `X` and can return any of the
            above. To select multiple columns by name or dtype, you can use
            :obj:`make_column_selector`.

    remainder : {'drop', 'passthrough'} or estimator, default='drop'
        By default, only the specified columns in `transformers` are
        transformed and combined in the output, and the non-specified
        columns are dropped. (default of ``'drop'``).
        By specifying ``remainder='passthrough'``, all remaining columns that
        were not specified in `transformers` will be automatically passed
        through. This subset of columns is concatenated with the output of
        the transformers.
        By setting ``remainder`` to be an estimator, the remaining
        non-specified columns will use the ``remainder`` estimator. The
        estimator must support ``fit`` and ``transform``.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        n_jobs does not work with with GPU backends.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.

    Returns
    -------
    column_transformer_no_stack : ColumnTransformerNoStack

    See also
    --------
    himalaya.ridge.ColumnTransformerNoStack : Class that allows combining the
        outputs of multiple transformer objects used on column subsets of the
        data into a single feature space.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from himalaya.ridge import make_column_transformer_no_stack
    >>> ct = make_column_transformer_no_stack(
    ...     (StandardScaler(), [0, 1, 2]),
    ...     (StandardScaler(), slice(3, 5)))
    >>> # Group separately the first three columns and the last two
    >>> # columns, creating two feature spaces.
    >>> Xs = ct.fit_transform(X)
    >>> print([Xi.shape for Xi in Xs])
    [(4, 3), (4, 2)]
    """
    # transformer_weights keyword is not passed through because the user
    # would need to know the automatically generated names of the transformers
    n_jobs = kwargs.pop('n_jobs', None)
    remainder = kwargs.pop('remainder', 'drop')
    verbose = kwargs.pop('verbose', False)
    if kwargs:
        raise TypeError('Unknown keyword arguments: "{}"'.format(
            list(kwargs.keys())[0]))
    transformer_list = _get_transformer_list(transformers)
    return ColumnTransformerNoStack(transformer_list, n_jobs=n_jobs,
                                    remainder=remainder, verbose=verbose)


def _get_transformer_list(estimators):
    """
    Construct (name, trans, column) tuples from list
    """
    transformers, columns = zip(*estimators)
    names, _ = zip(*_name_estimators(transformers))

    transformer_list = list(zip(names, transformers, columns))
    return transformer_list



from . import backend
from . import kernel_ridge
from . import lasso
from . import ridge
from . import progress_bar
from . import scoring
from . import utils
from . import viz

__version__ = '0.3.1'

__all__ = [
    backend,
    kernel_ridge,
    lasso,
    ridge,
    progress_bar,
    scoring,
    utils,
    viz,
]



try:
    import cupy
except ImportError as error:
    import sys
    if "pytest" in sys.modules:  # if run through pytest
        import pytest
        pytest.skip("Cupy not installed.")
    raise ImportError("Cupy not installed.") from error

###############################################################################


def apply_argmax(array, argmax, axis):
    """Apply precomputed argmax indices in multi dimension arrays

    array[np.argmax(array)] works fine in dimension 1, but not in higher ones.
    This function extends it to higher dimensions.

    Examples
    --------
    >>> import cupy
    >>> array = cupy.random.randn(10, 4, 8)
    >>> argmax = cupy.argmax(array, axis=1)
    >>> max_ = apply_argmax(array, argmax, axis=1)
    >>> assert cupy.all(max_ == cupy.max(array, axis=1))
    """
    argmax = cupy.expand_dims(argmax, axis=axis)
    max_ = cupy.take_along_axis(array, argmax, axis=axis)
    return cupy.take(max_, 0, axis=axis)


def std_float64(array, axis=None, demean=True, keepdims=False):
    """Compute the standard deviation of X with double precision,
    and cast back the result to original dtype.
    """
    return array.std(axis, dtype=cupy.float64,
                     keepdims=keepdims).astype(array.dtype, copy=False)


def mean_float64(array, axis=None, keepdims=False):
    """Compute the mean of X with double precision,
    and cast back the result to original dtype.
    """
    return array.mean(axis, dtype=cupy.float64,
                      keepdims=keepdims).astype(array.dtype, copy=False)


###############################################################################

name = "cupy"
argmax = cupy.argmax
max = cupy.max
min = cupy.min
abs = cupy.abs
randn = cupy.random.randn
rand = cupy.random.rand
matmul = cupy.matmul
transpose = cupy.transpose
stack = cupy.stack
concatenate = cupy.concatenate
sum = cupy.sum
sqrt = cupy.sqrt
any = cupy.any
all = cupy.all
nan = cupy.nan
inf = cupy.inf
isnan = cupy.isnan
isinf = cupy.isinf
logspace = cupy.logspace
copy = cupy.copy
bool = cupy.bool
float32 = cupy.float32
float64 = cupy.float64
int32 = cupy.int32
eigh = cupy.linalg.eigh
norm = cupy.linalg.norm
log = cupy.log
exp = cupy.exp
arange = cupy.arange
flatnonzero = cupy.flatnonzero
unique = cupy.unique
einsum = cupy.einsum
tanh = cupy.tanh
power = cupy.power
prod = cupy.prod
zeros = cupy.zeros
sign = cupy.sign
clip = cupy.clip
sort = cupy.sort
flip = cupy.flip
atleast_1d = cupy.atleast_1d
finfo = cupy.finfo


def diagonal_view(array, axis1=0, axis2=1):
    """Return a view of the array diagonal."""
    return cupy.diagonal(array, 0, axis1=axis1, axis2=axis2)


def to_numpy(array):
    return cupy.asnumpy(array)


def isin(x, y):
    import numpy as np  # XXX
    np_result = np.isin(to_numpy(x), to_numpy(y))
    return asarray(np_result, dtype=cupy.bool)


def searchsorted(x, y):
    import numpy as np  # XXX
    np_result = np.searchsorted(to_numpy(x), to_numpy(y))
    return asarray(np_result, dtype=cupy.int64)


def zeros_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in zeros_like."""
    xp = cupy.get_array_module(array)
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    if device == "cpu":
        import numpy as xp
    return xp.zeros(shape, dtype=dtype)


def ones_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in ones_like."""
    xp = cupy.get_array_module(array)
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    if device == "cpu":
        import numpy as xp
    return xp.ones(shape, dtype=dtype)


def full_like(array, fill_value, shape=None, dtype=None, device=None):
    """Add a shape parameter in full_like."""
    xp = cupy.get_array_module(array)
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    if device == "cpu":
        import numpy as xp
    return xp.full(shape, fill_value, dtype=dtype)


def to_cpu(array):
    return cupy.asnumpy(array)


def to_gpu(array, device=None):
    return cupy.asarray(array)


def asarray(a, dtype=None, order=None, device=None):
    if device == "cpu":
        import numpy as np
        return np.asarray(cupy.asnumpy(a), dtype, order)
    else:
        return cupy.asarray(a, dtype, order)


def asarray_like(x, ref):
    xp = cupy.get_array_module(ref)
    return xp.asarray(x, dtype=ref.dtype)


def check_arrays(*all_inputs):
    """Change all inputs into arrays (or list of arrays) using the same
    precision as the first one. Some arrays can be None.
    """
    all_arrays = []
    all_arrays.append(asarray(all_inputs[0]))
    dtype = all_arrays[0].dtype
    for tensor in all_inputs[1:]:
        if tensor is None:
            pass
        elif isinstance(tensor, list):
            tensor = [asarray(tt, dtype=dtype) for tt in tensor]
        else:
            tensor = asarray(tensor, dtype=dtype)
        all_arrays.append(tensor)
    return all_arrays


def svd(X, full_matrices=True):
    if X.ndim == 2:
        return cupy.linalg.svd(X, full_matrices=full_matrices)
    elif X.ndim == 3:
        UsV_list = [
            cupy.linalg.svd(Xi, full_matrices=full_matrices) for Xi in X
        ]
        return map(cupy.stack, zip(*UsV_list))
    else:
        raise NotImplementedError()



"""
Multiple kernel ridge with scikit-learn API
===========================================
This example demonstrates how to solve multiple kernel ridge regression, using
scikit-learn API.
"""

import numpy as np
import matplotlib.pyplot as plt

from himalaya.backend import set_backend
from himalaya.kernel_ridge import KernelRidgeCV
from himalaya.kernel_ridge import MultipleKernelRidgeCV
from himalaya.kernel_ridge import Kernelizer
from himalaya.kernel_ridge import ColumnKernelizer

from sklearn.pipeline import make_pipeline

# sphinx_gallery_thumbnail_number = 2
###############################################################################
# In this example, we use the ``torch`` backend.
#
# Torch can perform computations both on CPU and GPU.
# To use the CPU, use the "torch" backend.
# To use GPU, you can either use the "torch" backend and move your data to GPU
# with the ``.cuda`` method, or you can use the "torch_cuda" backend which calls
# this method in ``backend.asarray``.

backend = set_backend("torch_cuda")

###############################################################################
# Generate a random dataset
# -------------------------
# - Xs_train : list of arrays of shape (n_samples_train, n_features)
# - Xs_test : list of arrays of shape (n_samples_test, n_features)
# - Y_train : array of shape (n_samples_train, n_targets)
# - Y_test : array of shape (n_repeat, n_samples_test, n_targets)

n_samples_train = 1000
n_samples_test = 300
n_targets = 1000
n_features_list = [1000, 1000, 500]
feature_names = ["feature space A", "feature space B", "feature space C"]

Xs_train = [
    backend.randn(n_samples_train, n_features)
    for n_features in n_features_list
]
Xs_test = [
    backend.randn(n_samples_test, n_features) for n_features in n_features_list
]
ws = [
    backend.randn(n_features, n_targets) / n_features
    for n_features in n_features_list
]
Y_train = backend.stack([X @ w for X, w in zip(Xs_train, ws)]).sum(0)
Y_test = backend.stack([X @ w for X, w in zip(Xs_test, ws)]).sum(0)

###############################################################################
# Optional: Add some arbitrary scalings per kernel
if True:
    scalings = [0.2, 5, 1]
    Xs_train = [X * scaling for X, scaling in zip(Xs_train, scalings)]
    Xs_test = [X * scaling for X, scaling in zip(Xs_test, scalings)]

###############################################################################
# Concatenate the feature spaces and move to GPU with ``backend.asarray``.
X_train = backend.asarray(backend.concatenate(Xs_train, 1), dtype="float32")
X_test = backend.asarray(backend.concatenate(Xs_test, 1), dtype="float32")

###############################################################################
# We could precompute the kernels by hand on ``Xs_train``, as done in
# ``plot_mkr_random_search.py``. Instead, here we use the
# ``ColumnKernelizer`` to make a ``scikit-learn`` ``Pipeline``.

# Find the start and end of each feature space X in Xs
start_and_end = np.concatenate([[0], np.cumsum(n_features_list)])
slices = [
    slice(start, end)
    for start, end in zip(start_and_end[:-1], start_and_end[1:])
]

###############################################################################
# Create a different ``Kernelizer`` for each feature space. Here we use a
# linear kernel for all feature spaces, but ``ColumnKernelizer`` accepts any
# ``Kernelizer``, or ``scikit-learn`` ``Pipeline`` ending with a
# ``Kernelizer``.
kernelizers = [(name, Kernelizer(), slice_)
               for name, slice_ in zip(feature_names, slices)]
column_kernelizer = ColumnKernelizer(kernelizers)

# Note that ``ColumnKernelizer`` has a parameter ``n_jobs`` to parallelize each
# kernelizer, yet such parallelism does not work with GPU arrays.

###############################################################################
# Define the model
# ----------------
#
# The class takes a number of common parameters during initialization, such as
# `kernels` or `solver`. Since the solver parameters might be different
# depending on the solver, they can be passed in the `solver_params` parameter.

###############################################################################
# Here we use the "random_search" solver.
# We can check its specific parameters in the function docstring:
solver_function = MultipleKernelRidgeCV.ALL_SOLVERS["random_search"]
print("Docstring of the function %s:" % solver_function.__name__)
print(solver_function.__doc__)

###############################################################################
# We use 100 iterations to have a reasonably fast example (~40 sec).
# To have a better convergence, we probably need more iterations.
# Note that there is currently no stopping criterion in this method.
n_iter = 100

###############################################################################
# Grid of regularization parameters.
alphas = np.logspace(-10, 10, 41)

###############################################################################
# Batch parameters are used to reduce the necessary GPU memory. A larger value
# will be a bit faster, but the solver might crash if it runs out of memory.
# Optimal values depend on the size of your dataset.
n_targets_batch = 1000
n_alphas_batch = 20
n_targets_batch_refit = 200

solver_params = dict(n_iter=n_iter, alphas=alphas,
                     n_targets_batch=n_targets_batch,
                     n_alphas_batch=n_alphas_batch,
                     n_targets_batch_refit=n_targets_batch_refit,
                     jitter_alphas=True)

model = MultipleKernelRidgeCV(kernels="precomputed", solver="random_search",
                              solver_params=solver_params)

###############################################################################
# Define and fit the pipeline
pipe = make_pipeline(column_kernelizer, model)
pipe.fit(X_train, Y_train)

###############################################################################
# Plot the convergence curve
# --------------------------

# ``cv_scores`` gives the scores for each sampled kernel weights.
# The convergence curve is thus the current maximum for each target.
cv_scores = backend.to_numpy(pipe[1].cv_scores_)
current_max = np.maximum.accumulate(cv_scores, axis=0)
mean_current_max = np.mean(current_max, axis=1)

x_array = np.arange(1, len(mean_current_max) + 1)
plt.plot(x_array, mean_current_max, '-o')
plt.grid("on")
plt.xlabel("Number of kernel weights sampled")
plt.ylabel("L2 negative loss (higher is better)")
plt.title("Convergence curve, averaged over targets")
plt.tight_layout()
plt.show()

###############################################################################
# Compare to ``KernelRidgeCV``
# ----------------------------
# Compare to a baseline ``KernelRidgeCV`` model with all the concatenated features.
# Comparison is performed using the prediction scores on the test set.


###############################################################################
# Fit the baseline model ``KernelRidgeCV``
baseline = KernelRidgeCV(kernel="linear", alphas=alphas)
baseline.fit(X_train, Y_train)

###############################################################################
# Compute scores of both models
scores = pipe.score(X_test, Y_test)
scores = backend.to_numpy(scores)

scores_baseline = baseline.score(X_test, Y_test)
scores_baseline = backend.to_numpy(scores_baseline)

###############################################################################
# Plot histograms
bins = np.linspace(min(scores_baseline.min(), scores.min()),
                   max(scores_baseline.max(), scores.max()), 50)
plt.hist(scores, bins, alpha=0.5, label="MultipleKernelRidgeCV")
plt.hist(scores_baseline, bins, alpha=0.5, label="KernelRidgeCV")
plt.xlabel(r"$R^2$ generalization score")
plt.title("Histogram over targets")
plt.legend()
plt.show()



from himalaya.progress_bar import bar
from himalaya.progress_bar import ProgressBar


def test_progress_bar():
    # simple smoke test
    for ii in bar(range(10), title="La barre"):
        pass

    bar_ = ProgressBar(title="La barre", max_value=10, initial_value=0,
                       max_chars=40, progress_character='.', spinner=False,
                       verbose_bool=True)
    for ii in bar_(range(10)):
        pass

    bar_ = ProgressBar(max_value=10, title="La barre")
    for ii in range(10):
        bar_.update_with_increment_value(1)
    bar_.close()



import itertools
import pytest

import sklearn.linear_model
import sklearn.model_selection

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS

from himalaya.kernel_ridge import solve_multiple_kernel_ridge_random_search
from himalaya.kernel_ridge import solve_multiple_kernel_ridge_hyper_gradient


def _create_dataset(backend):
    n_featuress = (50, 80)
    n_samples = 30
    n_targets = 2
    n_gammas = 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in n_featuress
    ]
    Ks = backend.stack([X @ X.T for X in Xs])

    ws = [
        backend.asarray(backend.randn(n_features, n_targets), backend.float64)
        for n_features in n_featuress
    ]
    Ys = backend.stack([X @ w for X, w in zip(Xs, ws)])
    Y = Ys.sum(0)

    gammas = backend.asarray(backend.rand(n_gammas, Ks.shape[0]),
                             backend.float64)
    gammas /= gammas.sum(1)[:, None]

    return Ks, Y, gammas


@pytest.mark.parametrize('Ks_in_cpu', [True, False])
@pytest.mark.parametrize('Y_in_cpu', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_random_search(backend, Ks_in_cpu, Y_in_cpu):
    backend = set_backend(backend)

    Ks, Y, gammas = _create_dataset(backend)
    gammas = gammas[:1]
    alphas = backend.asarray_like(backend.logspace(-3, 5, 3), Ks)
    cv = sklearn.model_selection.check_cv(2)

    for Ks_, Y_, gammas_, alphas_ in itertools.product(
        [Ks, backend.to_numpy(Ks),
         backend.to_cpu(Ks)],
        [Y, backend.to_numpy(Y), backend.to_cpu(Y)],
        [gammas, backend.to_numpy(gammas),
         backend.to_cpu(gammas), 2],
        [alphas, backend.to_numpy(alphas),
         backend.to_cpu(alphas)],
    ):

        deltas, _, _ = solve_multiple_kernel_ridge_random_search(
            Ks_, Y_, n_iter=gammas_, alphas=alphas_, cv=cv, progress_bar=False,
            Ks_in_cpu=Ks_in_cpu, Y_in_cpu=Y_in_cpu)

        assert deltas.dtype == Ks.dtype
        assert getattr(deltas, "device", None) == getattr(Ks, "device", None)


@pytest.mark.parametrize('Y_in_cpu', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_hyper_gradient(backend, Y_in_cpu):
    backend = set_backend(backend)

    Ks, Y, _ = _create_dataset(backend)
    cv = sklearn.model_selection.check_cv(2)

    for Ks_, Y_ in itertools.product(
        [Ks, backend.to_numpy(Ks),
         backend.to_cpu(Ks)],
        [Y, backend.to_numpy(Y), backend.to_cpu(Y)],
    ):

        deltas, _, _ = solve_multiple_kernel_ridge_hyper_gradient(
            Ks_, Y_, max_iter=1, cv=cv, progress_bar=False, Y_in_cpu=Y_in_cpu)

        assert deltas.dtype == Ks.dtype
        assert getattr(deltas, "device", None) == getattr(Ks, "device", None)



import pytest
import sklearn.utils.estimator_checks
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.exceptions import NotFittedError
from sklearn.base import clone

from himalaya.backend import set_backend
from himalaya.backend import get_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.kernel_ridge import Kernelizer
from himalaya.kernel_ridge import ColumnKernelizer
from himalaya.kernel_ridge import make_column_kernelizer
from himalaya.kernel_ridge._kernelizer import _end_with_a_kernel
from himalaya.kernel_ridge import KernelRidge
from himalaya.kernel_ridge import MultipleKernelRidgeCV
from himalaya.kernel_ridge import KernelCenterer


@pytest.mark.parametrize('kernel', Kernelizer.ALL_KERNELS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernelizer(backend, kernel):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    function = Kernelizer.ALL_KERNELS[kernel]
    ker = Kernelizer(kernel=kernel)

    assert_array_almost_equal(function(X1), ker.fit_transform(X1))
    assert_array_almost_equal(function(X2, X1), ker.transform(X2))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernelizer_callable(backend):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    def _callable_linear_kernel(x, y):
        return x @ y

    function = Kernelizer.ALL_KERNELS["linear"]
    ker = Kernelizer(kernel=_callable_linear_kernel)

    assert_array_almost_equal(function(X1), ker.fit_transform(X1))
    assert_array_almost_equal(function(X2, X1), ker.transform(X2))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernelizer_wrong_input(backend):
    backend = set_backend(backend)
    X1 = backend.randn(10, 5)
    X2 = backend.randn(10, 4)

    with pytest.raises(ValueError, match="Unknown metric"):
        Kernelizer(kernel="wrong").fit(X1)

    with pytest.raises(ValueError, match="Different number of features"):
        Kernelizer().fit(X1).transform(X2)


@pytest.mark.parametrize('kernel', Kernelizer.ALL_KERNELS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_all_columns(backend, kernel):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    function = Kernelizer.ALL_KERNELS[kernel]
    ker = ColumnKernelizer([
        ("name", Kernelizer(kernel=kernel), slice(0, 5)),
    ])

    assert ker.fit_transform(X1).shape == (1, 10, 10)
    assert ker.transform(X2).shape == (1, 8, 10)

    assert_array_almost_equal(function(X1), ker.fit_transform(X1)[0])
    assert_array_almost_equal(function(X2, X1), ker.transform(X2)[0])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_passthrough(backend):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    function = Kernelizer.ALL_KERNELS["linear"]
    ker = ColumnKernelizer([
        ("name", "passthrough", slice(0, 5)),
    ])

    assert ker.fit_transform(X1).shape == (1, 10, 10)
    assert ker.transform(X2).shape == (1, 8, 10)

    assert_array_almost_equal(function(X1), ker.fit_transform(X1)[0])
    assert_array_almost_equal(function(X2, X1), ker.transform(X2)[0])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_remainder(backend):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    function = Kernelizer.ALL_KERNELS["linear"]
    ker = ColumnKernelizer([
        ("name", Kernelizer(kernel="wrong"), []),
    ], remainder="passthrough")

    assert ker.fit_transform(X1).shape == (1, 10, 10)
    assert ker.transform(X2).shape == (1, 8, 10)

    assert_array_almost_equal(function(X1), ker.fit_transform(X1)[0])
    assert_array_almost_equal(function(X2, X1), ker.transform(X2)[0])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_multiple(backend):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    linear = Kernelizer.ALL_KERNELS["linear"]
    poly = Kernelizer.ALL_KERNELS["poly"]
    ker = ColumnKernelizer([
        ("name0", Kernelizer(kernel="linear"), [0, 1]),
        ("name1", Kernelizer(kernel="poly"), [2, 3]),
    ], remainder="passthrough")

    assert ker.fit_transform(X1).shape == (3, 10, 10)
    assert ker.transform(X2).shape == (3, 8, 10)

    assert_array_almost_equal(linear(X1[:, :2]), ker.fit_transform(X1)[0])
    assert_array_almost_equal(poly(X1[:, 2:4]), ker.fit_transform(X1)[1], 5)
    assert_array_almost_equal(linear(X1[:, 4:]), ker.fit_transform(X1)[2])
    assert_array_almost_equal(linear(X2[:, :2], X1[:, :2]),
                              ker.transform(X2)[0])
    assert_array_almost_equal(poly(X2[:, 2:4], X1[:, 2:4]),
                              ker.transform(X2)[1], 5)
    assert_array_almost_equal(linear(X2[:, 4:], X1[:, 4:]),
                              ker.transform(X2)[2])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_pipeline(backend):
    backend = set_backend(backend)

    X1 = backend.randn(10, 5)
    X2 = backend.randn(8, 5)

    pipe = make_pipeline("passthrough", Kernelizer(kernel="poly"))

    function = Kernelizer.ALL_KERNELS["poly"]
    ker = ColumnKernelizer([
        ("name", pipe, slice(0, 4)),
    ])

    assert ker.fit_transform(X1).shape == (1, 10, 10)
    assert ker.transform(X2).shape == (1, 8, 10)

    assert_array_almost_equal(function(X1[:, :4]), ker.fit_transform(X1)[0])
    assert_array_almost_equal(function(X2[:, :4], X1[:, :4]),
                              ker.transform(X2)[0])


def test_column_kernelizer_end_with_a_kernel():
    with pytest.raises(ValueError):
        _end_with_a_kernel("passthrough")
    with pytest.raises(ValueError):
        _end_with_a_kernel(Kernelizer)

    assert _end_with_a_kernel(StandardScaler()) is False
    assert _end_with_a_kernel(KernelRidge()) is False
    assert _end_with_a_kernel(ColumnKernelizer([])) is False

    assert _end_with_a_kernel(Kernelizer()) is True

    pipe = make_pipeline(Kernelizer(), StandardScaler())
    assert _end_with_a_kernel(pipe) is False
    pipe = make_pipeline(make_pipeline(Kernelizer()), StandardScaler())
    assert _end_with_a_kernel(pipe) is False

    pipe = make_pipeline(StandardScaler(), Kernelizer())
    assert _end_with_a_kernel(pipe) is True
    pipe = make_pipeline(Kernelizer())
    assert _end_with_a_kernel(pipe) is True
    pipe = make_pipeline(Kernelizer())
    pipe = make_pipeline(StandardScaler(), make_pipeline(Kernelizer()))
    assert _end_with_a_kernel(pipe) is True


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_make_column_kernelizer(backend):
    backend = set_backend(backend)

    trans = Kernelizer(kernel="cosine")
    ck = make_column_kernelizer((trans, slice(0, 3)))

    assert isinstance(ck, ColumnKernelizer)
    assert len(ck.transformers) == 1
    assert len(ck.transformers[0]) == 3
    assert ck.transformers[0][0] == "kernelizer"
    assert ck.transformers[0][1] == trans
    assert ck.transformers[0][2] == slice(0, 3)

    trans = Kernelizer(kernel="cosine")
    ck = make_column_kernelizer((trans, slice(0, 3)), ("passthrough", [3, 4]))

    assert isinstance(ck, ColumnKernelizer)
    assert len(ck.transformers) == 2
    assert len(ck.transformers[0]) == 3
    assert len(ck.transformers[1]) == 3
    assert ck.transformers[0][0] == "kernelizer"
    assert ck.transformers[0][1] == trans
    assert ck.transformers[0][2] == slice(0, 3)
    assert ck.transformers[1][0] == "passthrough"
    assert ck.transformers[1][1] == "passthrough"
    assert ck.transformers[1][2] == [3, 4]


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_in_pipeline(backend):
    backend = set_backend(backend)

    X = backend.randn(10, 5)
    Y = backend.randn(10, 3)

    ck = make_column_kernelizer(
        (Kernelizer("linear"), slice(0, 4)),
        (Kernelizer("linear"), slice(4, 6)),
    )
    pipe = make_pipeline(
        ck,
        MultipleKernelRidgeCV(
            kernels="precomputed",
            solver_params=dict(n_iter=backend.ones_like(X, shape=(1, 2)),
                               progress_bar=False)),
    )
    pipe.fit(X, Y)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernelizer_get_X_fit(backend):
    backend = set_backend(backend)
    X = backend.randn(10, 5)

    estimator = Kernelizer("linear")
    with pytest.raises(NotFittedError):
        estimator.get_X_fit()

    K = estimator.fit_transform(X)
    X = estimator.get_X_fit()
    assert len(K) == len(X)  # same number of samples


@pytest.mark.parametrize('estimator', [
    make_column_kernelizer((Kernelizer("linear"), slice(0, 4))),
    make_column_kernelizer(
        (Kernelizer("linear"), slice(0, 4)),
        ("drop", slice(4, 6)),
        remainder='passthrough',
    ),
    make_column_kernelizer(
        ("passthrough", slice(0, 2)),
        ("passthrough", slice(0, 4)),
        ("passthrough", slice(4, 5)),
    ),
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_kernelizer_get_X_fit(estimator, backend):
    backend = set_backend(backend)
    X = backend.randn(10, 5)

    estimator = clone(estimator)  # pytest reuses the same object over backends
    with pytest.raises(NotFittedError):
        estimator.get_X_fit()

    Ks = estimator.fit_transform(X)
    Xs = estimator.get_X_fit()
    assert len(Ks) == len(Xs)  # same number of feature spaces
    for X, K in zip(Xs, Ks):
        assert X.shape[0] == K.shape[0]  # same number of samples


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_centerer(backend):
    backend = set_backend(backend)
    X = backend.randn(10, 5) + 1
    K = X @ X.T

    X_centered = X - X.mean(0)
    K_centered = X_centered @ X_centered.T
    with pytest.raises(AssertionError):
        assert_array_almost_equal(K, K_centered, decimal=5)

    centerer = KernelCenterer()
    K2 = centerer.fit_transform(K)
    assert_array_almost_equal(K2, K_centered, decimal=5)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_centerer_wrong_input(backend):
    backend = set_backend(backend)
    X = backend.randn(10, 5)

    centerer = KernelCenterer()
    with pytest.raises(ValueError, match="must be a square matrix"):
        centerer.fit_transform(X)


###############################################################################
# scikit-learn.utils.estimator_checks


class Kernelizer_(Kernelizer):
    """Cast transforms to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def fit_transform(self, X, y=None):
        backend = get_backend()
        return backend.to_numpy(super().fit_transform(X, y))

    def transform(self, X):
        backend = get_backend()
        return backend.to_numpy(super().transform(X))


@sklearn.utils.estimator_checks.parametrize_with_checks([
    Kernelizer_(),
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_check_estimator(estimator, check, backend):
    backend = set_backend(backend)
    check(estimator)



from ._column import ColumnTransformerNoStack
from ._column import make_column_transformer_no_stack
from ._random_search import solve_group_ridge_random_search
from ._random_search import solve_ridge_cv_svd
from ._random_search import GROUP_RIDGE_SOLVERS
from ._solvers import solve_ridge_svd
from ._solvers import RIDGE_SOLVERS
from ._sklearn_api import Ridge
from ._sklearn_api import RidgeCV
from ._sklearn_api import GroupRidgeCV

# alternative names - kept for backward compatibility
BandedRidgeCV = GroupRidgeCV
solve_banded_ridge_random_search = solve_group_ridge_random_search
BANDED_RIDGE_SOLVERS = GROUP_RIDGE_SOLVERS

__all__ = [
    # column transformers
    "ColumnTransformerNoStack",
    "make_column_transformer_no_stack",
    # group ridge solvers
    "solve_group_ridge_random_search",
    "GROUP_RIDGE_SOLVERS",
    # ridge solvers
    "solve_ridge_svd",
    "solve_ridge_cv_svd",
    "RIDGE_SOLVERS",
    # sklearn API
    "Ridge",
    "RidgeCV",
    "GroupRidgeCV",
]
